from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Literal, Type, Union

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import aws, pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.util import S3Url
from great_expectations.datasource.fluent import _PandasFilePathDatasource
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
    _check_config_substitutions_needed,
)
from great_expectations.datasource.fluent.data_connector import (
    S3DataConnector,
)
from great_expectations.datasource.fluent.interfaces import TestConnectionError
from great_expectations.datasource.fluent.pandas_datasource import PandasDatasourceError

if TYPE_CHECKING:
    from botocore.client import BaseClient

    from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset

logger = logging.getLogger(__name__)


class PandasS3DatasourceError(PandasDatasourceError):
    pass


@public_api
class PandasS3Datasource(_PandasFilePathDatasource):
    """
    PandasS3Datasource is a PandasDatasource that uses Amazon S3 as a data store.
    """

    # class attributes
    data_connector_type: ClassVar[Type[S3DataConnector]] = S3DataConnector
    # these fields should not be passed to the execution engine
    _EXTRA_EXCLUDED_EXEC_ENG_ARGS: ClassVar[set] = {
        "bucket",
        "boto3_options",
    }

    # instance attributes
    type: Literal["pandas_s3"] = "pandas_s3"

    # S3 specific attributes
    bucket: str
    boto3_options: Dict[str, Union[ConfigStr, Any]] = {}

    _s3_client: Union[BaseClient, None] = pydantic.PrivateAttr(default=None)

    def _get_s3_client(self) -> BaseClient:
        s3_client: Union[BaseClient, None] = self._s3_client
        if not s3_client:
            # Validate that "boto3" library was successfully imported and attempt to create "s3_client" handle.  # noqa: E501 # FIXME CoP
            if aws.boto3:
                _check_config_substitutions_needed(
                    self, self.boto3_options, raise_warning_if_provider_not_present=True
                )
                # pull in needed config substitutions using the `_config_provider`
                # The `FluentBaseModel.dict()` call will do the config substitution on the serialized dict if a `config_provider` is passed  # noqa: E501 # FIXME CoP
                boto3_options: dict = self.dict(config_provider=self._config_provider).get(
                    "boto3_options", {}
                )
                try:
                    s3_client = aws.boto3.client("s3", **boto3_options)
                except Exception as e:
                    # Failure to create "s3_client" is most likely due invalid "boto3_options" dictionary.  # noqa: E501 # FIXME CoP
                    raise PandasS3DatasourceError(  # noqa: TRY003 # FIXME CoP
                        f'Due to exception: "{type(e).__name__}:{e}", "s3_client" could not be created.'  # noqa: E501 # FIXME CoP
                    ) from e
            else:
                raise PandasS3DatasourceError(  # noqa: TRY003 # FIXME CoP
                    'Unable to create "PandasS3Datasource" due to missing boto3 dependency.'
                )

            self._s3_client = s3_client

        return s3_client

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the PandasS3Datasource.

        Args:
            test_assets: If assets have been passed to the PandasS3Datasource, whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """  # noqa: E501 # FIXME CoP
        try:
            _ = self._get_s3_client()
        except Exception as e:
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                f"Attempt to connect to datasource failed with the following error message: {e!s}"
            ) from e

        if self.assets and test_assets:
            for asset in self.assets:
                asset.test_connection()

    @override
    def _build_data_connector(
        self,
        data_asset: FileDataAsset,
        s3_prefix: str = "",
        s3_delimiter: str = "/",  # TODO: delimiter conflicts with csv asset args
        s3_max_keys: int = 1000,
        s3_recursive_file_discovery: bool = False,
        **kwargs,
    ) -> None:
        """Builds and attaches the `S3DataConnector` to the asset."""
        # TODO: use the `asset_options_type` for validation and defaults
        if kwargs:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"_build_data_connector() got unexpected keyword arguments {list(kwargs.keys())}"
            )

        data_asset._data_connector = self.data_connector_type.build_data_connector(
            datasource_name=self.name,
            data_asset_name=data_asset.name,
            s3_client=self._get_s3_client(),
            bucket=self.bucket,
            prefix=s3_prefix,
            delimiter=s3_delimiter,
            max_keys=s3_max_keys,
            recursive_file_discovery=s3_recursive_file_discovery,
            file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
        )

        # build a more specific `_test_connection_error_message`
        data_asset._test_connection_error_message = (
            self.data_connector_type.build_test_connection_error_message(
                data_asset_name=data_asset.name,
                bucket=self.bucket,
                prefix=s3_prefix,
                delimiter=s3_delimiter,
                recursive_file_discovery=s3_recursive_file_discovery,
            )
        )

        logger.info(f"{self.data_connector_type.__name__} created for '{data_asset.name}'")



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.compatibility import pydantic
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

STRING_COL = "string_col"
INT_COL = "int_col"
INT_COL_2 = "int_col_2"
DUPLICATES = "duplicates"

try:
    from great_expectations.compatibility.pyspark import types as PYSPARK_TYPES

    SPARK_COLUMN_TYPES = {
        STRING_COL: PYSPARK_TYPES.StringType,
        INT_COL: PYSPARK_TYPES.IntegerType,
        INT_COL_2: PYSPARK_TYPES.IntegerType,
        DUPLICATES: PYSPARK_TYPES.IntegerType,
    }
except ModuleNotFoundError:
    SPARK_COLUMN_TYPES = {}


DATA = pd.DataFrame(
    {
        STRING_COL: ["foo", "bar", "foo", "baz", None, None],
        INT_COL: [1, 2, 1, 3, None, None],
        INT_COL_2: [1, 2, 3, 4, None, None],
        DUPLICATES: [100, 100, 100, 100, 99, 99],
    }
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_golden_path(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectCompoundColumnsToBeUnique(
        column_list=[STRING_COL, INT_COL, INT_COL_2],
        ignore_row_if="any_value_is_missing",
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectCompoundColumnsToBeUnique(column_list=[INT_COL, INT_COL_2]),
            id="two_cols",
        ),
        pytest.param(
            gxe.ExpectCompoundColumnsToBeUnique(column_list=[STRING_COL, INT_COL, INT_COL_2]),
            id="three_cols",
        ),
        pytest.param(
            gxe.ExpectCompoundColumnsToBeUnique(column_list=[INT_COL, DUPLICATES], mostly=0.3),
            id="mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectCompoundColumnsToBeUnique
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectCompoundColumnsToBeUnique(column_list=[INT_COL, DUPLICATES]),
        ),
        pytest.param(
            gxe.ExpectCompoundColumnsToBeUnique(column_list=[INT_COL, DUPLICATES], mostly=0.4),
            id="mostly_threshold_not_met",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectCompoundColumnsToBeUnique
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.unit
@pytest.mark.parametrize(
    "column_list",
    [
        pytest.param([], id="no_cols"),
        pytest.param([INT_COL_2], id="one_col"),
    ],
)
def test_invalid_config(column_list: list[str]) -> None:
    with pytest.raises(pydantic.ValidationError):
        gxe.ExpectCompoundColumnsToBeUnique(column_list=column_list)



from __future__ import annotations

import re
import uuid
from datetime import date, datetime
from enum import Enum
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import dateutil
from typing_extensions import TypeAlias, TypedDict

from great_expectations.compatibility.pydantic import (
    BaseModel,
    Field,
    ValidationError,
    create_model,
    root_validator,
    validator,
)
from great_expectations.compatibility.pydantic import generics as pydantic_generics
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import (
    ExpectationValidationResult,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,  # noqa: TC001 # FIXME CoP
)
from great_expectations.render.exceptions import RendererConfigurationError
from great_expectations.render.renderer.observed_value_renderer import ObservedValueRenderState

if TYPE_CHECKING:
    from great_expectations.compatibility.pydantic.typing import (
        AbstractSetIntStr,
        DictStrAny,
        MappingIntStrAny,
    )


class RendererValueType(str, Enum):
    """Type used in renderer param json schema dictionary."""

    ARRAY = "array"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    NUMBER = "number"
    OBJECT = "object"
    STRING = "string"

    @classmethod
    def from_value(cls, value: Any) -> RendererValueType:  # noqa: PLR0911
        if value is None:
            return RendererValueType.STRING

        if isinstance(value, list):
            return RendererValueType.ARRAY
        elif isinstance(value, bool):
            return RendererValueType.BOOLEAN
        elif isinstance(value, (date, datetime)):
            return RendererValueType.DATETIME
        elif isinstance(value, Number):
            return RendererValueType.NUMBER
        elif isinstance(value, dict):
            return RendererValueType.OBJECT
        elif isinstance(value, (str, uuid.UUID)):
            return RendererValueType.STRING
        else:
            raise TypeError


class RendererSchema(TypedDict):
    """Json schema for values found in renderers."""

    type: RendererValueType


class _RendererValueBase(BaseModel):
    """
    _RendererValueBase is the base for renderer classes that need to override the default pydantic dict behavior.
    """  # noqa: E501 # FIXME CoP

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @override
    def dict(  # noqa: PLR0913 # FIXME CoP
        self,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = True,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
    ) -> DictStrAny:
        """
        Override BaseModel dict to make the defaults:
            - by_alias=True because we have an existing attribute named schema, and schema is already a Pydantic
              BaseModel attribute.
            - exclude_none=True to ensure that None values aren't included in the json dict.

        In practice this means the renderer implementer doesn't need to use .dict(by_alias=True, exclude_none=True)
        everywhere.
        """  # noqa: E501 # FIXME CoP
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )


RendererParams = TypeVar("RendererParams", bound=_RendererValueBase)

RendererValueTypes: TypeAlias = Union[RendererValueType, List[RendererValueType]]

AddParamArgs: TypeAlias = Tuple[Tuple[str, RendererValueTypes], ...]


class RendererTableValue(_RendererValueBase):
    """Represents each value within a row of a header_row or a table."""

    renderer_schema: RendererSchema = Field(alias="schema")
    value: Optional[Any]


class MetaNotesFormat(str, Enum):
    """Possible formats that can be rendered via MetaNotes."""

    STRING = "string"
    MARKDOWN = "markdown"


class MetaNotes(TypedDict):
    """Notes that can be added to the meta field of an Expectation."""

    format: MetaNotesFormat
    content: List[str]


class CodeBlockLanguage(str, Enum):
    JSON = "json"
    PYTHON = "python"
    SQL = "sql"
    YAML = "yaml"


class CodeBlock(TypedDict):
    """A code block of a specified language to be rendered.
    The code_template_str uses $variable substitution of RendererConfiguration.params.
    e.g. RendererConfiguration.params: python_code_block_1, python_code_block_2

    CodeBlock(
        code_template_str="$python_code_block_1\n\n$python_code_block_2",
        language=CodeBlockLanguage.PYTHON,
    )
    """

    code_template_str: str
    language: CodeBlockLanguage


def make_exception_msg(v_type: RendererValueType, input: Any) -> str:
    return f"Param type: <{v_type}> does not match value: <{input}>."


def make_str_exception_msg(v_type: RendererValueType, input: Any) -> str:
    message = f"Value was unable to be represented as a {v_type}"
    try:
        str(input)
    except Exception as e:
        message = f"{message}: {e}"
    return message


def safe_str(input: Any) -> bool:
    try:
        str(input)
        return True
    except Exception:
        return False


def safe_date(input: Any) -> bool:
    try:
        dateutil.parser.parse(input)
        return True
    except Exception:
        return False


PARAM_TYPE_TESTERS: Dict[
    RendererValueType,
    Tuple[
        Callable[[Any], bool],
        Callable[[RendererValueType, Any], str],
    ],
] = {
    RendererValueType.ARRAY: (
        lambda x: isinstance(x, (list, tuple, set)),
        make_exception_msg,
    ),
    RendererValueType.BOOLEAN: (
        lambda x: isinstance(x, bool),
        make_exception_msg,
    ),
    RendererValueType.DATETIME: (
        lambda x: isinstance(x, (date, datetime)) or safe_date(x),
        make_exception_msg,
    ),
    RendererValueType.NUMBER: (
        lambda x: isinstance(x, Number),
        make_exception_msg,
    ),
    RendererValueType.OBJECT: (lambda x: True, make_exception_msg),
    RendererValueType.STRING: (
        lambda x: safe_str(x),
        make_str_exception_msg,
    ),
}


class RendererConfiguration(pydantic_generics.GenericModel, Generic[RendererParams]):
    """
    Configuration object built for each renderer. Operations to be performed strictly on this object at the renderer
        implementation-level.
    """  # noqa: E501 # FIXME CoP

    configuration: Optional[ExpectationConfiguration] = Field(None, allow_mutation=False)
    result: Optional[ExpectationValidationResult] = Field(None, allow_mutation=False)
    runtime_configuration: Optional[dict] = Field({}, allow_mutation=False)
    expectation_type: str = Field("", allow_mutation=False)
    kwargs: dict = Field({}, allow_mutation=False)
    meta_notes: MetaNotes = Field(
        MetaNotes(format=MetaNotesFormat.STRING, content=[]), allow_mutation=False
    )
    template_str: Optional[str] = Field(None, allow_mutation=True)
    code_block: CodeBlock = Field({}, allow_mutation=True)
    header_row: List[RendererTableValue] = Field([], allow_mutation=True)
    table: List[List[RendererTableValue]] = Field([], allow_mutation=True)
    graph: dict = Field({}, allow_mutation=True)
    include_column_name: bool = Field(True, allow_mutation=False)
    _raw_kwargs: dict = Field({}, allow_mutation=False)
    _row_condition: str = Field("", allow_mutation=False)
    params: RendererParams = Field(..., allow_mutation=True)

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    # TODO: Reintroduce this constraint once legacy renderers are deprecated/removed
    # @root_validator(pre=True)
    def _validate_configuration_or_result(cls, values: dict) -> dict:
        if ("configuration" not in values or values["configuration"] is None) and (
            "result" not in values or values["result"] is None
        ):
            raise RendererConfigurationError(  # noqa: TRY003 # FIXME CoP
                "RendererConfiguration must be passed either configuration or result."
            )
        return values

    def __init__(self, **values) -> None:
        values["params"] = _RendererValueBase()
        super().__init__(**values)

    class _RequiredRendererParamArgs(TypedDict):
        """Used for building up a dictionary that is unpacked into RendererParams upon initialization."""  # noqa: E501 # FIXME CoP

        schema: RendererSchema
        value: Any

    class _RendererParamArgs(_RequiredRendererParamArgs, total=False):
        """Used for building up a dictionary that is unpacked into RendererParams upon initialization."""  # noqa: E501 # FIXME CoP

        suite_parameter: Dict[str, Any]

    class _RendererParamBase(_RendererValueBase):
        """
        _RendererParamBase is the base for a param that is added to RendererParams. It contains the validation logic,
            but it is dynamically renamed in order for the RendererParams attribute to have the same name as the param.
        """  # noqa: E501 # FIXME CoP

        renderer_schema: RendererSchema = Field(alias="schema", allow_mutation=False)
        value: Any = Field(allow_mutation=False)
        suite_parameter: Optional[Dict[str, Any]] = Field(allow_mutation=False)
        render_state: Optional[ObservedValueRenderState] = Field(allow_mutation=True)

        class Config:
            validate_assignment = True
            arbitrary_types_allowed = True

        @root_validator(pre=True)
        def _validate_param_type_matches_value(cls, values: dict) -> dict:
            """
            This root_validator ensures that a value can be parsed by its RendererValueType.
            If RendererValueType.OBJECT is passed, it is treated as valid for any value.
            """
            # if render_state is in values, this root_validator is being called
            # on assignment rather than during instantiation
            if not values.get("render_state"):
                param_type: RendererValueType = values["schema"]["type"]
                value: Any = values["value"]

                (tester, error) = PARAM_TYPE_TESTERS[param_type]
                if not tester(value):
                    raise RendererConfigurationError(error(param_type, value))
            return values

        @override
        def __eq__(self, other: object) -> bool:
            if isinstance(other, BaseModel):
                return self.dict() == other.dict()
            elif isinstance(other, dict):
                return self.dict() == other
            else:
                return NotImplemented

        @override
        def __hash__(self) -> int:
            return hash(tuple(sorted(self.dict().items())))

    @staticmethod
    def _get_renderer_value_base_model_type(
        name: str,
    ) -> Type[BaseModel]:
        return create_model(
            name,
            renderer_schema=(
                RendererSchema,
                Field(..., alias="schema"),
            ),
            value=(Union[Any, None], ...),
            __base__=RendererConfiguration._RendererParamBase,
        )

    @staticmethod
    def _get_suite_parameter_params_from_raw_kwargs(
        raw_kwargs: Dict[str, Any],
    ) -> Dict[str, RendererConfiguration._RendererParamArgs]:
        renderer_params_args = {}
        for kwarg_name, value in raw_kwargs.items():
            renderer_params_args[kwarg_name] = RendererConfiguration._RendererParamArgs(
                schema=RendererSchema(type=RendererValueType.OBJECT),
                value=None,
                suite_parameter={
                    "schema": RendererSchema(type=RendererValueType.OBJECT),
                    "value": value,
                },
            )
        return renderer_params_args

    @root_validator()
    def _validate_and_set_renderer_attrs(cls, values: dict) -> dict:
        if (
            "result" in values
            and values["result"] is not None
            and values["result"].expectation_config is not None
        ):
            expectation_configuration: ExpectationConfiguration = values[
                "result"
            ].expectation_config
            values["expectation_type"] = expectation_configuration.type
            values["kwargs"] = expectation_configuration.kwargs
            raw_configuration: ExpectationConfiguration = (
                expectation_configuration.get_raw_configuration()
            )
            if "_raw_kwargs" not in values:
                values["_raw_kwargs"] = {
                    key: value
                    for key, value in raw_configuration.kwargs.items()
                    if (key, value) not in values["kwargs"].items()
                }
                renderer_params_args: Dict[str, RendererConfiguration._RendererParamArgs] = (
                    RendererConfiguration._get_suite_parameter_params_from_raw_kwargs(
                        raw_kwargs=values["_raw_kwargs"]
                    )
                )
                values["_params"] = (
                    {**values["_params"], **renderer_params_args}
                    if values.get("_params")
                    else renderer_params_args
                )
        elif "configuration" in values and values["configuration"] is not None:
            values["expectation_type"] = values["configuration"].type
            values["kwargs"] = values["configuration"].kwargs
            # description is the template_str override
            if values["configuration"].description:
                values["template_str"] = values["configuration"].description

        return values

    @root_validator()
    def _validate_for_include_column_name(cls, values: dict) -> dict:
        if values.get("runtime_configuration"):
            values["include_column_name"] = (
                values["runtime_configuration"].get("include_column_name") is not False
            )
        return values

    @staticmethod
    def _get_row_condition_params(
        row_condition_str: str,
    ) -> Dict[str, RendererConfiguration._RendererParamArgs]:
        row_condition_str = RendererConfiguration._parse_row_condition_str(
            row_condition_str=row_condition_str
        )
        row_conditions_list: List[str] = (
            RendererConfiguration._get_row_conditions_list_from_row_condition_str(
                row_condition_str=row_condition_str
            )
        )
        renderer_params_args = {}
        for idx, condition in enumerate(row_conditions_list):
            name = f"row_condition__{idx!s}"
            value = condition.replace(" NOT ", " not ")
            renderer_params_args[name] = RendererConfiguration._RendererParamArgs(
                schema=RendererSchema(type=RendererValueType.STRING), value=value
            )
        return renderer_params_args

    @root_validator()
    def _validate_for_row_condition(cls, values: dict) -> dict:
        kwargs: Dict[str, Any]
        if (
            "result" in values
            and values["result"] is not None
            and values["result"].expectation_config is not None
        ):
            kwargs = values["result"].expectation_config.kwargs
        else:
            kwargs = values["configuration"].kwargs

        values["_row_condition"] = kwargs.get("row_condition", "")
        if values["_row_condition"]:
            renderer_params_args: Dict[str, RendererConfiguration._RendererParamArgs] = (
                RendererConfiguration._get_row_condition_params(
                    row_condition_str=values["_row_condition"],
                )
            )
            values["_params"] = (
                {**values["_params"], **renderer_params_args}
                if values.get("_params")
                else renderer_params_args
            )
        return values

    @root_validator()
    def _validate_for_meta_notes(cls, values: dict) -> dict:
        meta_notes: Optional[dict[str, Optional[dict[str, list[str] | tuple[str] | str]]]]
        if (
            "result" in values
            and values["result"] is not None
            and values["result"].expectation_config is not None
        ):
            meta_notes = values["result"].expectation_config.meta.get("notes")
        else:
            meta_notes = values["configuration"].meta.get("notes")

        if meta_notes and isinstance(meta_notes, dict):
            meta_notes_content = meta_notes.get("content")

            if isinstance(meta_notes_content, (list, tuple)):
                meta_notes["content"] = list(meta_notes_content)
            elif isinstance(meta_notes_content, str):
                meta_notes["content"] = [meta_notes_content]
            values["meta_notes"] = meta_notes
        elif meta_notes and isinstance(meta_notes, str):
            values["meta_notes"] = {
                "content": [meta_notes],
                "format": MetaNotesFormat.STRING,
            }

        return values

    @root_validator()
    def _validate_for_params(cls, values: dict) -> dict:
        _params: Optional[Dict[str, Dict[str, Union[str, Dict[str, RendererValueType]]]]] = (
            values.get("_params")
        )
        # this logic loads the params that exist
        # upon RendererConfiguration instantiation, e.g. row_condition
        if _params and len(values["params"].__dict__) == 0:
            renderer_param_definitions: Dict[str, Any] = {}
            for name in _params:
                renderer_param_type: Type[BaseModel] = (
                    RendererConfiguration._get_renderer_value_base_model_type(name=name)
                )
                renderer_param_definitions[name] = (
                    Optional[renderer_param_type],
                    ...,
                )
            renderer_params: Type[BaseModel] = create_model(
                "RendererParams",
                **renderer_param_definitions,
                __base__=_RendererValueBase,
            )
            values["params"] = renderer_params(**_params)

        return values

    @staticmethod
    def _get_row_conditions_list_from_row_condition_str(
        row_condition_str: str,
    ) -> List[str]:
        # extract the variables to be substituted
        row_conditions_list = re.split(r"AND|OR|NOT(?! in)|col\(\"|\"\)", row_condition_str)
        row_conditions_list = [
            condition.strip() for condition in row_conditions_list if condition.strip()
        ]
        return row_conditions_list

    @staticmethod
    def _parse_row_condition_str(row_condition_str: str) -> str:
        if not row_condition_str:
            row_condition_str = "True"

        row_condition_str = (
            row_condition_str.replace("&", " AND ")
            .replace(" and ", " AND ")
            .replace("|", " OR ")
            .replace(" or ", " OR ")
            .replace("~", " NOT ")
            .replace(" not ", " NOT ")
        )
        row_condition_str = " ".join(row_condition_str.split())

        # replace tuples of values by lists of values
        tuples_list = re.findall(r"\([^()]*,[^()]*\)", row_condition_str)
        for value_tuple in tuples_list:
            value_list = value_tuple.replace("(", "[").replace(")", "]")
            row_condition_str = row_condition_str.replace(value_tuple, value_list)

        return row_condition_str

    @staticmethod
    def _get_row_condition_string(row_condition_str: str) -> str:
        row_condition_str = RendererConfiguration._parse_row_condition_str(
            row_condition_str=row_condition_str
        )
        row_conditions_list: List[str] = (
            RendererConfiguration._get_row_conditions_list_from_row_condition_str(
                row_condition_str=row_condition_str
            )
        )
        for idx, condition in enumerate(row_conditions_list):
            row_condition_str = row_condition_str.replace(condition, f"$row_condition__{idx!s}")
        row_condition_str = row_condition_str.lower()
        return f"If {row_condition_str}, then "

    @validator("template_str")
    def _set_template_str(cls, v: str, values: dict) -> str:
        if values.get("_row_condition"):
            row_condition_str: str = RendererConfiguration._get_row_condition_string(
                row_condition_str=values["_row_condition"]
            )
            v = row_condition_str + v

        return v

    @staticmethod
    def _choose_param_type_for_value(
        param_types: List[RendererValueType], value: Any
    ) -> RendererValueType:
        for param_type in param_types:
            try:
                renderer_param: Type[BaseModel] = (
                    RendererConfiguration._get_renderer_value_base_model_type(name="try_param")
                )
                renderer_param(schema=RendererSchema(type=param_type), value=value)
                return param_type
            except ValidationError:
                pass

        raise RendererConfigurationError(  # noqa: TRY003 # FIXME CoP
            f"None of the param_types: {[param_type.value for param_type in param_types]} match the value: {value}"  # noqa: E501 # FIXME CoP
        )

    def add_param(
        self,
        name: str,
        param_type: Optional[RendererValueTypes] = None,
        value: Optional[Any] = None,
    ) -> None:
        """Adds a param that can be substituted into a template string during rendering.

        Attributes:
            name (str): A name for the attribute to be added to the params of this RendererConfiguration instance.
            param_type (one or a list of RendererValueTypes): The possible types for the value being substituted. If
                zero or more than one param_type is passed, inference based on param value will be performed,
                and the first param_type to match the value will be selected.
                    One of:
                     - array
                     - boolean
                     - date
                     - number
                     - string
            value (Optional[Any]): The value to be substituted into the template string. If no value is
                provided, a value lookup will be attempted in RendererConfiguration.kwargs using the
                provided name.

        Returns:
            None
        """  # noqa: E501 # FIXME CoP
        renderer_param: Type[BaseModel] = RendererConfiguration._get_renderer_value_base_model_type(
            name=name
        )

        if value is None:
            value = self.kwargs.get(name)

        if param_type is None:
            param_type = sorted(
                RendererValueType,
                key=lambda x: (
                    # in order to infer type correctly
                    # object must be last in the list
                    # as it is permissive to any value
                    x.value == "object",
                    # and string must be second to last
                    # as it is permissive to string-able value
                    x.value == "string",
                    x.value,
                ),
            )

        if isinstance(value, dict) and "$PARAMETER" in value:
            param_type = RendererValueType.OBJECT
        elif isinstance(param_type, list) and value is not None:
            param_type = RendererConfiguration._choose_param_type_for_value(
                param_types=param_type, value=value
            )

        if value is None:
            self.params.__dict__[name] = None
            return

        assert isinstance(param_type, RendererValueType)

        # if we already moved the suite parameter raw_kwargs to a param,
        # we need to combine the param passed to add_param() with those existing raw_kwargs
        if name in self.params.__dict__ and self.params.__dict__[name]["suite_parameter"]:
            self.params.__dict__[name] = renderer_param(
                schema=RendererSchema(type=param_type),
                value=value,
                suite_parameter=self.params.__dict__[name]["suite_parameter"],
            )
        else:
            self.params.__dict__[name] = renderer_param(
                schema=RendererSchema(type=param_type),
                value=value,
            )



import pytest

from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.store import DatasourceStore
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)


@pytest.fixture
def datasource_store_ge_cloud_backend(
    ge_cloud_base_url: str,
    ge_cloud_access_token: str,
    ge_cloud_organization_id: str,
    datasource_store_name: str,
):
    ge_cloud_store_backend_config: dict = {
        "class_name": GXCloudStoreBackend.__name__,
        "ge_cloud_base_url": ge_cloud_base_url,
        "ge_cloud_resource_type": GXCloudRESTResource.DATASOURCE,
        "ge_cloud_credentials": {
            "access_token": ge_cloud_access_token,
            "organization_id": ge_cloud_organization_id,
        },
        "suppress_store_backend_id": True,
    }

    store = DatasourceStore(
        store_name=datasource_store_name,
        store_backend=ge_cloud_store_backend_config,
    )
    return store



from __future__ import annotations

import enum
import os
import pathlib
import subprocess
import uuid
from typing import TYPE_CHECKING, Any, Callable, Dict, Final, List, Union

import pact
import pytest
from typing_extensions import Annotated, TypeAlias

from great_expectations.compatibility import pydantic
from great_expectations.core.http import create_session
from great_expectations.data_context import CloudDataContext
from great_expectations.data_context.data_context.context_factory import project_manager

if TYPE_CHECKING:
    from requests import Session


CONSUMER_NAME: Final[str] = "great_expectations"
PROVIDER_NAME: Final[str] = "mercury"


PACT_MOCK_HOST: Final[str] = "localhost"
PACT_MOCK_PORT: Final[int] = 9292
PACT_DIR: Final[pathlib.Path] = pathlib.Path(pathlib.Path(__file__, ".."), "pacts").resolve()
PACT_MOCK_SERVICE_URL: Final[str] = f"http://{PACT_MOCK_HOST}:{PACT_MOCK_PORT}"


JsonData: TypeAlias = Union[None, int, str, bool, List[Any], Dict[str, Any]]

PactBody: TypeAlias = Union[
    Dict[str, Union[JsonData, pact.matchers.Matcher]], pact.matchers.Matcher, None
]


EXISTING_ORGANIZATION_ID: Final[str] = os.environ.get("GX_CLOUD_ORGANIZATION_ID", "")


class RequestMethods(str, enum.Enum):
    DELETE = "DELETE"
    GET = "GET"
    PATCH = "PATCH"
    POST = "POST"
    PUT = "PUT"


@pytest.fixture
def cloud_base_url() -> str:
    try:
        return os.environ["GX_CLOUD_BASE_URL"]
    except KeyError as e:
        raise OSError("GX_CLOUD_BASE_URL is not set in this environment.") from e


@pytest.fixture
def cloud_access_token() -> str:
    try:
        return os.environ["GX_CLOUD_ACCESS_TOKEN"]
    except KeyError as e:
        raise OSError("GX_CLOUD_ACCESS_TOKEN is not set in this environment.") from e


@pytest.fixture(scope="module")
def gx_cloud_session() -> Session:
    try:
        access_token = os.environ["GX_CLOUD_ACCESS_TOKEN"]
    except KeyError as e:
        raise OSError("GX_CLOUD_ACCESS_TOKEN is not set in this environment.") from e
    return create_session(access_token=access_token)


@pytest.fixture
def cloud_data_context(
    cloud_base_url: str,
    cloud_access_token: str,
) -> CloudDataContext:
    """This is a real Cloud Data Context that points to the pact mock service instead of the Mercury API."""  # noqa: E501 # FIXME CoP
    cloud_data_context = CloudDataContext(
        cloud_base_url=cloud_base_url,
        cloud_organization_id=EXISTING_ORGANIZATION_ID,
        cloud_access_token=cloud_access_token,
    )
    # we can't override the base url to use the mock service due to
    # reliance on env vars, so instead we override with a real project config
    project_config = cloud_data_context.config
    context = CloudDataContext(
        cloud_base_url=PACT_MOCK_SERVICE_URL,
        cloud_organization_id=EXISTING_ORGANIZATION_ID,
        cloud_access_token=cloud_access_token,
        project_config=project_config,
    )
    project_manager.set_project(cloud_data_context)
    return context


def get_git_commit_hash() -> str:
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()


@pytest.fixture(scope="package")
def pact_test(request) -> pact.Pact:
    """
    pact_test can be used as a context manager and will:
    1. write a new contract to the pact dir
    2. verify the contract against the mock service
    """
    pact_broker_base_url = "https://greatexpectations.pactflow.io"

    broker_token: str
    publish_to_broker: bool
    if os.environ.get("PACT_BROKER_READ_WRITE_TOKEN"):
        broker_token = os.environ.get("PACT_BROKER_READ_WRITE_TOKEN", "")
        # do not publish to broker on develop until we have integrated the 1.0 API with GX Cloud
        publish_to_broker = False
    elif os.environ.get("PACT_BROKER_READ_ONLY_TOKEN"):
        broker_token = os.environ.get("PACT_BROKER_READ_ONLY_TOKEN", "")
        publish_to_broker = False
    else:
        pytest.skip(
            "no pact credentials: set PACT_BROKER_READ_ONLY_TOKEN from greatexpectations.pactflow.io"  # noqa: E501 # FIXME CoP
        )

    # Adding random id to the commit hash allows us to run the build
    # and publish the contract more than once for a given commit.
    # We need this because we have the ability to trigger re-run of tests
    # in GH, and we run the release build process on the tagged commit.
    version = f"{get_git_commit_hash()}_{str(uuid.uuid4())[:5]}"

    _pact: pact.Pact = pact.Consumer(
        name=CONSUMER_NAME,
        version=version,
        tag_with_git_branch=True,
        auto_detect_version_properties=True,
    ).has_pact_with(
        pact.Provider(name=PROVIDER_NAME),
        broker_base_url=pact_broker_base_url,
        broker_token=broker_token,
        host_name=PACT_MOCK_HOST,
        port=PACT_MOCK_PORT,
        pact_dir=str(PACT_DIR),
        publish_to_broker=publish_to_broker,
    )

    _pact.start_service()
    yield _pact
    _pact.stop_service()


class ContractInteraction(pydantic.BaseModel):
    """Represents a Python API (Consumer) request and expected minimal response,
       given a state in the Cloud backend (Provider).

    The given state is something you know to be true about the Cloud backend data requested.

    Args:
        method: A string (e.g. "GET" or "POST") or attribute of the RequestMethods class representing a request method.
        request_path: A pathlib.Path to the endpoint relative to the base url.
                        e.g.
                            ```
                            path = pathlib.Path(
                                "/", "organizations", organization_id, "data-context-configuration"
                            )
                            ```
        upon_receiving: A string description of the type of request being made.
        given: A string description of the state of the Cloud backend data requested.
        response_status: The status code associated with the response. An integer between 100 and 599.
        response_body: A dictionary or Pact Matcher object representing the response body.
        request_body (Optional): A dictionary or Pact Matcher object representing the request body.
        request_headers (Optional): A dictionary representing the request headers.
        request_parmas (Optional): A dictionary representing the request parameters.

    Returns:
        ContractInteraction
    """  # noqa: E501 # FIXME CoP

    class Config:
        arbitrary_types_allowed = True

    method: Union[RequestMethods, pydantic.StrictStr]
    request_path: pathlib.Path
    upon_receiving: pydantic.StrictStr
    given: pydantic.StrictStr
    response_status: Annotated[int, pydantic.Field(strict=True, ge=100, lt=600)]
    response_body: PactBody
    request_body: Union[PactBody, None] = None
    request_headers: Union[dict, None] = None
    request_params: Union[dict, None] = None


@pytest.fixture
def run_rest_api_pact_test(
    gx_cloud_session: Session,
    pact_test: pact.Pact,
) -> Callable:
    def _run_pact_test(
        contract_interaction: ContractInteraction,
    ) -> None:
        """Runs a contract test and produces a Pact contract json file in directory:
            - tests/integration/cloud/rest_contracts/pacts

        Args:
            contract_interaction: A ContractInteraction object which represents a Python API (Consumer) request
                                  and expected minimal response, given a state in the Cloud backend (Provider).

        Returns:
            None
        """  # noqa: E501 # FIXME CoP

        request: dict[str, str | PactBody] = {
            "method": contract_interaction.method,
            "path": str(contract_interaction.request_path),
        }
        if contract_interaction.request_body is not None:
            request["body"] = contract_interaction.request_body
        if contract_interaction.request_params is not None:
            request["query"] = contract_interaction.request_params

        request["headers"] = dict(gx_cloud_session.headers)
        if contract_interaction.request_headers is not None:
            request["headers"].update(contract_interaction.request_headers)  # type: ignore[union-attr] # FIXME CoP
            gx_cloud_session.headers.update(contract_interaction.request_headers)

        response: dict[str, int | PactBody] = {
            "status": contract_interaction.response_status,
        }
        if contract_interaction.response_body is not None:
            response["body"] = contract_interaction.response_body

        (
            pact_test.given(provider_state=contract_interaction.given)
            .upon_receiving(scenario=contract_interaction.upon_receiving)
            .with_request(**request)
            .will_respond_with(**response)
        )

        request_url = f"http://{PACT_MOCK_HOST}:{PACT_MOCK_PORT}{contract_interaction.request_path}"

        with pact_test:
            # act
            resp = gx_cloud_session.request(
                method=contract_interaction.method,
                url=request_url,
                json=contract_interaction.request_body,
                params=contract_interaction.request_params,
            )

        # assert
        assert resp.status_code == contract_interaction.response_status
        # TODO more unit test assertions would go here e.g. response body checks

    return _run_pact_test



from __future__ import annotations

import json
import pathlib
import uuid
from typing import TYPE_CHECKING, Generator, List, Literal, Type
from unittest import mock

import pandas as pd
import pytest
from requests import Session

import great_expectations as gx
from great_expectations import expectations as gxe
from great_expectations.analytics.events import CheckpointRanEvent
from great_expectations.checkpoint import (
    MicrosoftTeamsNotificationAction,
    SlackNotificationAction,
    UpdateDataDocsAction,
)
from great_expectations.checkpoint.actions import ValidationAction
from great_expectations.checkpoint.checkpoint import (
    Checkpoint,
    CheckpointResult,
)
from great_expectations.compatibility.pydantic import ValidationError
from great_expectations.constants import DATAFRAME_REPLACEMENT_STR
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.freshness_diagnostics import (
    BatchDefinitionFreshnessDiagnostics,
    CheckpointFreshnessDiagnostics,
    ExpectationSuiteFreshnessDiagnostics,
    ValidationDefinitionFreshnessDiagnostics,
)
from great_expectations.core.result_format import ResultFormat
from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.data_context.data_context.context_factory import set_context
from great_expectations.data_context.data_context.ephemeral_data_context import (
    EphemeralDataContext,
)
from great_expectations.data_context.types.resource_identifiers import (
    ValidationResultIdentifier,
)
from great_expectations.exceptions import (
    BatchDefinitionNotAddedError,
    CheckpointNotAddedError,
    CheckpointRelatedResourcesFreshnessError,
    CheckpointRunWithoutValidationDefinitionError,
    ExpectationSuiteNotAddedError,
    ResourceFreshnessError,
    ValidationDefinitionNotAddedError,
)
from great_expectations.exceptions.exceptions import (
    CheckpointNotFoundError,
    ValidationActionRegistryRetrievalError,
    ValidationDefinitionNotFoundError,
)
from great_expectations.exceptions.resource_freshness import ResourceFreshnessAggregateError
from great_expectations.expectations.expectation_configuration import ExpectationConfiguration
from great_expectations.expectations.window import Offset, Window
from tests.test_utils import working_directory

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from great_expectations.data_context.data_context.cloud_data_context import CloudDataContext


@pytest.mark.unit
def test_checkpoint_no_validation_definitions_does_not_raise():
    Checkpoint(name="my_checkpoint", validation_definitions=[])
    # see, no errors!


@pytest.mark.unit
def test_checkpoint_save_success(mocker: MockerFixture):
    context = mocker.Mock(spec=AbstractDataContext)
    set_context(project=context)

    checkpoint = Checkpoint(
        name="my_checkpoint",
        validation_definitions=[mocker.Mock(spec=ValidationDefinition)],
    )
    store_key = context.checkpoint_store.get_key.return_value
    checkpoint.save()

    context.checkpoint_store.update.assert_called_once_with(key=store_key, value=checkpoint)


class TestValidationDefinitionInteraction:
    """
    These tests are specifically to verify the following workflow:
    * Users instantiates (but doesn't save) a ValidationDefinition
    * User instantiates (but doesn't save) a Checkpoint with the ValidationDefinition
    * User saves the ValidationDefinition
    * User saves the Checkpoint

    See https://greatexpectations.atlassian.net/browse/V1-496.
    """

    @pytest.mark.unit
    def test_checkpoint_does_not_clone_validation_definition(self):
        context = gx.get_context(mode="ephemeral")

        batch_definition = (
            context.data_sources.add_pandas("my_datasource")
            .add_dataframe_asset("my_asset")
            .add_batch_definition_whole_dataframe("my_batch_definition")
        )
        suite = context.suites.add(ExpectationSuite(name="my_suite"))
        validation_definition = ValidationDefinition(
            name="my_validation_definition", suite=suite, data=batch_definition
        )

        checkpoint = Checkpoint(
            name="my_checkpoint",
            validation_definitions=[validation_definition],
        )

        assert checkpoint.validation_definitions[0] is validation_definition

    @pytest.mark.filesystem
    def test_save_order_does_not_matter(self, tmp_path: pathlib.Path):
        with working_directory(tmp_path):
            context = gx.get_context(mode="file")

            batch_definition = (
                context.data_sources.add_pandas("my_datasource")
                .add_dataframe_asset("my_asset")
                .add_batch_definition_whole_dataframe("my_batch_definition")
            )
            suite = context.suites.add(ExpectationSuite(name="my_suite"))
            validation_definition = ValidationDefinition(
                name="my_validation_definition", suite=suite, data=batch_definition
            )

            checkpoint = Checkpoint(
                name="my_checkpoint",
                validation_definitions=[validation_definition],
            )

            context.validation_definitions.add(validation_definition)
            context.checkpoints.add(checkpoint)


@pytest.fixture
def slack_action():
    return SlackNotificationAction(
        name="my_slack_action",
        slack_webhook="${SLACK_WEBHOOK}",
    )


@pytest.fixture
def teams_action():
    return MicrosoftTeamsNotificationAction(
        name="my_teams_action",
        teams_webhook="teams_webhook",
    )


@pytest.fixture
def actions(
    slack_action: SlackNotificationAction,
    teams_action: MicrosoftTeamsNotificationAction,
) -> list[ValidationAction]:
    return [slack_action, teams_action]


class TestCheckpointSerialization:
    @pytest.fixture
    def in_memory_context(self) -> EphemeralDataContext:
        return gx.get_context(mode="ephemeral")

    @pytest.fixture
    def validation_definition_1(
        self, in_memory_context: EphemeralDataContext, mocker: MockerFixture
    ):
        name = "my_first_validation"
        vc = ValidationDefinition(
            name=name,
            data=mocker.Mock(spec=BatchDefinition),
            suite=mocker.Mock(spec=ExpectationSuite),
        )
        with (
            mock.patch.object(
                ValidationDefinition,
                "json",
                return_value=json.dumps({"id": str(uuid.uuid4()), "name": name}),
            ),
            mock.patch.object(
                ValidationDefinition,
                "is_fresh",
                return_value=ValidationDefinitionFreshnessDiagnostics(errors=[]),
            ),
        ):
            yield in_memory_context.validation_definitions.add(vc)

    @pytest.fixture
    def validation_definition_2(
        self, in_memory_context: EphemeralDataContext, mocker: MockerFixture
    ):
        name = "my_second_validation"
        vc = ValidationDefinition(
            name=name,
            data=mocker.Mock(spec=BatchDefinition),
            suite=mocker.Mock(spec=ExpectationSuite),
        )
        with (
            mock.patch.object(
                ValidationDefinition,
                "json",
                return_value=json.dumps({"id": str(uuid.uuid4()), "name": name}),
            ),
            mock.patch.object(
                ValidationDefinition,
                "is_fresh",
                return_value=ValidationDefinitionFreshnessDiagnostics(errors=[]),
            ),
        ):
            yield in_memory_context.validation_definitions.add(vc)

    @pytest.fixture
    def validation_definitions(
        self,
        validation_definition_1: ValidationDefinition,
        validation_definition_2: ValidationDefinition,
    ) -> list[ValidationDefinition]:
        return [validation_definition_1, validation_definition_2]

    @pytest.mark.parametrize(
        "action_fixture_name, expected_actions",
        [
            pytest.param(None, [], id="no_actions"),
            pytest.param(
                "actions",
                [
                    {
                        "name": "my_slack_action",
                        "notify_on": "all",
                        "notify_with": None,
                        "renderer": {
                            "class_name": "SlackRenderer",
                            "module_name": "great_expectations.render.renderer.slack_renderer",
                        },
                        "show_failed_expectations": False,
                        "slack_channel": None,
                        "slack_token": None,
                        "slack_webhook": "${SLACK_WEBHOOK}",
                        "type": "slack",
                    },
                    {
                        "name": "my_teams_action",
                        "notify_on": "all",
                        "renderer": {
                            "class_name": "MicrosoftTeamsRenderer",
                            "module_name": "great_expectations.render.renderer.microsoft_teams_renderer",  # noqa: E501 # FIXME CoP
                        },
                        "teams_webhook": "teams_webhook",
                        "type": "microsoft",
                    },
                ],
                id="actions",
            ),
        ],
    )
    @pytest.mark.unit
    def test_checkpoint_serialization(
        self,
        validation_definitions: list[ValidationDefinition],
        action_fixture_name: str | None,
        expected_actions: dict,
        request: pytest.FixtureRequest,
    ):
        actions = request.getfixturevalue(action_fixture_name) if action_fixture_name else []
        cp = Checkpoint(
            name="my_checkpoint",
            validation_definitions=validation_definitions,
            actions=actions,
        )

        actual = json.loads(cp.json(models_as_dict=False))
        expected = {
            "name": cp.name,
            "validation_definitions": [
                {
                    "id": mock.ANY,
                    "name": "my_first_validation",
                },
                {
                    "id": mock.ANY,
                    "name": "my_second_validation",
                },
            ],
            "actions": expected_actions,
            "result_format": ResultFormat.SUMMARY,
            "id": cp.id,
        }

        assert actual == expected

        # Validation definitions should be persisted and obtain IDs before serialization
        self._assert_valid_uuid(actual["validation_definitions"][0]["id"])
        self._assert_valid_uuid(actual["validation_definitions"][1]["id"])

    @pytest.mark.filesystem
    def test_checkpoint_filesystem_round_trip_adds_ids(
        self,
        tmp_path: pathlib.Path,
        actions: list[ValidationAction],
    ):
        with working_directory(tmp_path):
            context = gx.get_context(mode="file")

        ds_name = "my_datasource"
        asset_name = "my_asset"
        batch_definition_name_1 = "my_batch1"
        suite_name_1 = "my_suite1"
        validation_definition_name_1 = "my_validation1"
        batch_definition_name_2 = "my_batch2"
        suite_name_2 = "my_suite2"
        validation_definition_name_2 = "my_validation2"
        cp_name = "my_checkpoint"

        ds = context.data_sources.add_pandas(ds_name)
        asset = ds.add_csv_asset(asset_name, "my_file.csv")  # type: ignore[arg-type] # FIXME CoP

        bc1 = asset.add_batch_definition(batch_definition_name_1)
        suite1 = context.suites.add(ExpectationSuite(suite_name_1))
        vc1 = context.validation_definitions.add(
            ValidationDefinition(name=validation_definition_name_1, data=bc1, suite=suite1)
        )

        bc2 = asset.add_batch_definition(batch_definition_name_2)
        suite2 = context.suites.add(ExpectationSuite(suite_name_2))
        vc2 = context.validation_definitions.add(
            ValidationDefinition(name=validation_definition_name_2, data=bc2, suite=suite2)
        )

        validation_definitions = [vc1, vc2]
        cp = Checkpoint(
            name=cp_name, validation_definitions=validation_definitions, actions=actions
        )

        serialized_checkpoint = cp.json(models_as_dict=False)
        serialized_checkpoint_dict = json.loads(serialized_checkpoint)

        assert serialized_checkpoint_dict == {
            "name": cp_name,
            "validation_definitions": [
                {
                    "id": mock.ANY,
                    "name": validation_definition_name_1,
                },
                {
                    "id": mock.ANY,
                    "name": validation_definition_name_2,
                },
            ],
            "actions": [
                {
                    "name": "my_slack_action",
                    "notify_on": "all",
                    "notify_with": None,
                    "renderer": {
                        "class_name": "SlackRenderer",
                        "module_name": "great_expectations.render.renderer.slack_renderer",
                    },
                    "show_failed_expectations": False,
                    "slack_channel": None,
                    "slack_token": None,
                    "slack_webhook": "${SLACK_WEBHOOK}",
                    "type": "slack",
                },
                {
                    "name": "my_teams_action",
                    "notify_on": "all",
                    "renderer": {
                        "class_name": "MicrosoftTeamsRenderer",
                        "module_name": "great_expectations.render.renderer.microsoft_teams_renderer",  # noqa: E501 # FIXME CoP
                    },
                    "teams_webhook": "teams_webhook",
                    "type": "microsoft",
                },
            ],
            "result_format": ResultFormat.SUMMARY,
            "id": None,
        }

        cp = Checkpoint.parse_raw(serialized_checkpoint)

        # Check that all nested objects have been built properly with their appropriate names
        assert cp.name == cp_name
        assert cp.validation_definitions[0].data_source.name == ds_name
        assert cp.validation_definitions[0].asset.name == asset_name

        assert cp.validation_definitions[0].name == validation_definition_name_1
        assert cp.validation_definitions[0].batch_definition.name == batch_definition_name_1
        assert cp.validation_definitions[0].suite.name == suite_name_1

        assert cp.validation_definitions[1].name == validation_definition_name_2
        assert cp.validation_definitions[1].batch_definition.name == batch_definition_name_2
        assert cp.validation_definitions[1].suite.name == suite_name_2

        # Check that all validation_definitions and nested suites have been assigned IDs during serialization  # noqa: E501 # FIXME CoP
        self._assert_valid_uuid(id=cp.validation_definitions[0].id)
        self._assert_valid_uuid(id=cp.validation_definitions[1].id)
        self._assert_valid_uuid(id=cp.validation_definitions[0].suite.id)
        self._assert_valid_uuid(id=cp.validation_definitions[1].suite.id)

    def _assert_valid_uuid(self, id: str | None) -> None:
        if not id:
            pytest.fail("id is None when it should be a valid UUID generated from a Store.")

        try:
            uuid.UUID(id)
        except ValueError:
            pytest.fail(f"{id} is not a valid UUID.")


class TestCheckpointDeserialization:
    @pytest.mark.parametrize(
        "serialized_checkpoint, expected_error",
        [
            pytest.param(
                {
                    "name": "my_checkpoint",
                    "validation_definitions": [
                        {
                            "name": "i_do_not_exist",
                            "id": "a758816-64c8-46cb-8f7e-03c12cea1d67",
                        }
                    ],
                    "actions": [],
                    "id": "c758816-64c8-46cb-8f7e-03c12cea1d67",
                },
                "Unable to retrieve validation definition",
                id="nonexistent_validation",
            ),
            pytest.param(
                {
                    "name": "my_checkpoint",
                    "validation_definitions": [
                        {
                            "other_key": "i_do_not_exist",
                            "id": "a758816-64c8-46cb-8f7e-03c12cea1d67",
                        }
                    ],
                    "actions": [],
                    "id": "c758816-64c8-46cb-8f7e-03c12cea1d67",
                },
                "validation_definitions -> name\n  field required",
                id="invalid_validation",
            ),
        ],
    )
    @pytest.mark.unit
    def test_checkpoint_deserialization_failure(
        self, serialized_checkpoint: str, expected_error: str
    ):
        with pytest.raises(ValidationError) as e:
            Checkpoint.parse_obj(serialized_checkpoint)

        assert expected_error in str(e.value)

    @pytest.fixture
    def _set_context(self, mocker: MockerFixture) -> Generator[None, None, None]:
        context = mocker.Mock(spec=AbstractDataContext)
        context.validation_definition_store.get.return_value = mocker.Mock(
            spec=ValidationDefinition
        )

        set_context(context)
        yield
        set_context(None)

    @pytest.mark.unit
    def test_checkpoint_deserialization_with_actions_success(
        self, _set_context: Generator[None, None, None]
    ):
        # Arrange
        serialized_checkpoint = {
            "actions": [
                {"name": "my_docs_action", "site_names": [], "type": "update_data_docs"},
                {"name": "my_slack_action", "slack_webhook": "test", "type": "slack"},
                {"name": "my_teams_action", "teams_webhook": "test", "type": "microsoft"},
            ],
            "id": "e7d1f462-821b-429c-8086-cca80eeea5e9",
            "name": "my_checkpoint",
            "result_format": "SUMMARY",
            "validation_definitions": [
                {"id": "3fb9ce09-a8fb-44d6-8abd-7d699443f6a1", "name": "my_validation_def"}
            ],
        }

        # Act
        checkpoint = Checkpoint.parse_obj(serialized_checkpoint)

        # Assert
        assert len(checkpoint.actions) == 3
        assert isinstance(checkpoint.actions[0], UpdateDataDocsAction)
        assert isinstance(checkpoint.actions[1], SlackNotificationAction)
        assert isinstance(checkpoint.actions[2], MicrosoftTeamsNotificationAction)

    @pytest.mark.parametrize(
        "action_config, expected_error",
        [
            pytest.param(
                {"name": "my_docs_action", "site_names": []},
                ValidationActionRegistryRetrievalError,
                id="no_type",
            ),
            pytest.param(
                {"name": "my_custom_action", "type": "not_registered"},
                ValidationActionRegistryRetrievalError,
                id="not_registered",
            ),
        ],
    )
    @pytest.mark.unit
    def test_checkpoint_deserialization_with_actions_failure(
        self,
        _set_context: Generator[None, None, None],
        action_config: dict,
        expected_error: Type[Exception],
    ):
        # Arrange
        serialized_checkpoint = {
            "actions": [
                action_config,
            ],
            "id": "e7d1f462-821b-429c-8086-cca80eeea5e9",
            "name": "my_checkpoint",
            "validation_definitions": [
                {"id": "3fb9ce09-a8fb-44d6-8abd-7d699443f6a1", "name": "my_validation_def"}
            ],
        }

        # Act & Assert
        with pytest.raises(expected_error):
            Checkpoint.parse_obj(serialized_checkpoint)

    @pytest.mark.unit
    def test_checkpoint_deserialization_with_custom_validation_action(
        self, _set_context: Generator[None, None, None]
    ):
        # Arrange
        class CustomAction(ValidationAction):
            type: Literal["custom"] = "custom"

        serialized_checkpoint = {
            "actions": [
                {"name": "my_custom_action", "type": "custom"},
            ],
            "id": "e7d1f462-821b-429c-8086-cca80eeea5e9",
            "name": "my_checkpoint",
            "validation_definitions": [
                {"id": "3fb9ce09-a8fb-44d6-8abd-7d699443f6a1", "name": "my_validation_def"}
            ],
        }

        # Act
        checkpoint = Checkpoint.parse_obj(serialized_checkpoint)

        # Assert
        assert isinstance(checkpoint.actions[0], CustomAction)


class TestCheckpointResult:
    suite_name: str = "my_suite"
    column_name = "passenger_count"
    datasource_name: str = "my_pandas_datasource"
    asset_name: str = "my_asset"
    batch_definition_name: str = "my_batch_def"
    validation_definition_name: str = "my_validation_def"
    checkpoint_name: str = "my_checkpoint"

    @pytest.fixture
    def mock_suite(self, mocker: MockerFixture):
        suite = mocker.Mock(spec=ExpectationSuite)
        suite.name = self.suite_name
        suite.is_fresh.return_value = ExpectationSuiteFreshnessDiagnostics(errors=[])
        return suite

    @pytest.fixture
    def mock_batch_def(self, mocker: MockerFixture):
        bd = mocker.Mock(spec=BatchDefinition)
        bd._copy_and_set_values().is_fresh.return_value = BatchDefinitionFreshnessDiagnostics(
            errors=[]
        )
        return bd

    @pytest.fixture
    def validation_definition(self, mock_suite: MockerFixture, mock_batch_def: MockerFixture):
        validation_definition = ValidationDefinition(
            name=self.validation_definition_name,
            id=str(uuid.uuid4()),
            data=mock_batch_def,
            suite=mock_suite,
        )

        mock_run_result = ExpectationSuiteValidationResult(
            success=True,
            results=[
                ExpectationValidationResult(
                    success=True,
                    expectation_config=ExpectationConfiguration(
                        type="expect_column_values_to_be_between",
                        kwargs={
                            "batch_id": f"{self.datasource_name}-{self.asset_name}",
                            "column": self.column_name,
                            "min_value": 0.0,
                            "max_value": 100.0,
                        },
                    ),
                    result={
                        "element_count": 100000,
                        "unexpected_count": 0,
                        "unexpected_percent": 0.0,
                        "partial_unexpected_list": [],
                        "missing_count": 0,
                        "missing_percent": 0.0,
                        "unexpected_percent_total": 0.0,
                        "unexpected_percent_nonmissing": 0.0,
                        "partial_unexpected_counts": [],
                        "partial_unexpected_index_list": [],
                    },
                ),
            ],
            suite_name=self.suite_name,
            statistics={
                "evaluated_expectations": 1,
                "successful_expectations": 1,
                "unsuccessful_expectations": 0,
                "success_percent": 100.0,
            },
            batch_id=f"{self.datasource_name}-{self.asset_name}",
        )

        with mock.patch.object(ValidationDefinition, "run", return_value=mock_run_result):
            yield validation_definition

    @pytest.mark.unit
    def test_checkpoint_run_no_validation_definitions(self, mocker):
        context = mocker.Mock(spec=AbstractDataContext)
        set_context(project=context)
        checkpoint = Checkpoint(name=self.checkpoint_name, validation_definitions=[])

        with pytest.raises(CheckpointRunWithoutValidationDefinitionError):
            checkpoint.run()

    @pytest.mark.unit
    def test_checkpoint_run_dependencies_not_added_raises_error(
        self, validation_definition: ValidationDefinition
    ):
        validation_definition.id = None
        checkpoint = Checkpoint(
            name=self.checkpoint_name, validation_definitions=[validation_definition]
        )

        with pytest.raises(CheckpointRelatedResourcesFreshnessError) as e:
            checkpoint.run()

        assert [type(err) for err in e.value.errors] == [
            ValidationDefinitionNotAddedError,
            CheckpointNotAddedError,
        ]

    @pytest.mark.unit
    def test_checkpoint_run_no_actions(
        self, validation_definition: ValidationDefinition, mocker: MockerFixture
    ):
        context = mocker.Mock(spec=AbstractDataContext)
        set_context(project=context)
        checkpoint = Checkpoint(
            name=self.checkpoint_name, validation_definitions=[validation_definition]
        )

        result = checkpoint.run()

        run_results = result.run_results
        assert len(run_results) == len(checkpoint.validation_definitions)

        validation_result = tuple(run_results.values())[0]
        assert validation_result.success is True
        assert len(validation_result.results) == 1 and validation_result.results[0].success is True

        assert result.success is True

    @pytest.mark.unit
    def test_checkpoint_run_actions(
        self,
        validation_definition: ValidationDefinition,
        mocker: MockerFixture,
    ):
        # Arrange
        action = mocker.Mock(spec=UpdateDataDocsAction, type="update_data_docs")
        checkpoint = Checkpoint(
            name=self.checkpoint_name,
            validation_definitions=[validation_definition],
            actions=[action],
        )

        # Act
        result = checkpoint.run()

        # Assert
        action._copy_and_set_values().run.assert_called_once_with(
            checkpoint_result=result, action_context=mock.ANY
        )

    @pytest.mark.unit
    def test_checkpoint_sorts_actions(self, validation_definition: ValidationDefinition):
        """
        Note that we are directly testing the `_sort_actions()` private method here.

        This was done as a way to expedite the testing process due to some conflicts with
        Pydantics and mocks.
        Ideally, this would be tested through the public `run()` method.
        """
        slack_action = SlackNotificationAction(
            name="my_slack_action", slack_webhook="${SLACK_WEBHOOK}"
        )
        teams_action = MicrosoftTeamsNotificationAction(
            name="my_teams_action", teams_webhook="teams_webhook"
        )
        data_docs_action = UpdateDataDocsAction(name="my_docs_action")

        actions: List[ValidationAction] = [slack_action, teams_action, data_docs_action]

        validation_definitions = [validation_definition]
        checkpoint = Checkpoint(
            name=self.checkpoint_name,
            validation_definitions=validation_definitions,
            actions=actions,
        )

        assert checkpoint._sort_actions() == [data_docs_action, slack_action, teams_action]

    @pytest.mark.unit
    def test_checkpoint_run_passes_through_runtime_params(
        self, validation_definition: ValidationDefinition
    ):
        checkpoint_id = str(uuid.uuid4())
        checkpoint = Checkpoint(
            id=checkpoint_id,
            name=self.checkpoint_name,
            validation_definitions=[validation_definition],
        )
        batch_parameters = {"my_param": "my_value"}
        expectation_parameters = {"my_other_param": "my_other_value"}

        with mock.patch.object(
            Checkpoint, "is_fresh", return_value=CheckpointFreshnessDiagnostics(errors=[])
        ):
            _ = checkpoint.run(
                batch_parameters=batch_parameters, expectation_parameters=expectation_parameters
            )

        validation_definition.run.assert_called_with(  # type: ignore[attr-defined] # FIXME CoP
            checkpoint_id=checkpoint_id,
            batch_parameters=batch_parameters,
            expectation_parameters=expectation_parameters,
            result_format=ResultFormat.SUMMARY,
            run_id=mock.ANY,
        )

    @pytest.mark.unit
    def test_checkpoint_run_sends_analytics(
        self, validation_definition: ValidationDefinition, mocker: MockerFixture
    ):
        # Arrange
        submit_analytics_event = mocker.patch(
            "great_expectations.checkpoint.checkpoint.submit_analytics_event"
        )

        context = mocker.Mock(spec=AbstractDataContext)
        set_context(project=context)
        checkpoint_id = "e051d0a8-d492-4cde-91f3-29f353758488"
        checkpoint = Checkpoint(
            id=checkpoint_id,
            name=self.checkpoint_name,
            validation_definitions=[validation_definition],
        )

        # Act
        with mock.patch.object(
            Checkpoint, "is_fresh", return_value=CheckpointFreshnessDiagnostics(errors=[])
        ):
            checkpoint.run()

        # Assert
        submit_analytics_event.assert_called_once_with(
            event=CheckpointRanEvent(
                checkpoint_id=checkpoint_id, validation_definition_ids=[validation_definition.id]
            )
        )

    @pytest.mark.unit
    def test_result_init_no_run_results_raises_error(self, mocker: MockerFixture):
        with pytest.raises(ValueError) as e:
            CheckpointResult(
                run_id=mocker.Mock(spec=RunIdentifier),
                run_results={},
                checkpoint_config=mocker.Mock(spec=Checkpoint),
            )

        assert "CheckpointResult must contain at least one run result" in str(e.value)

    @pytest.mark.unit
    def test_result_describe(self, mocker: MockerFixture):
        result = CheckpointResult(
            run_id=mocker.Mock(spec=RunIdentifier),
            run_results={
                mocker.Mock(spec=ValidationResultIdentifier): mocker.Mock(
                    spec=ExpectationSuiteValidationResult, success=True
                ),
            },
            checkpoint_config=mocker.Mock(spec=Checkpoint),
        )

        with mock.patch.object(
            ExpectationSuiteValidationResult,
            "describe_dict",
            return_value={
                "success": True,
                "statistics": {
                    "evaluated_expectations": 1,
                    "successful_expectations": 1,
                    "unsuccessful_expectations": 0,
                    "success_percent": 100.0,
                },
                "expectations": [
                    {
                        "expectation_type": "expect_column_values_to_be_between",
                        "success": True,
                        "kwargs": {
                            "batch_id": "default_pandas_datasource-#ephemeral_pandas_asset",
                            "mostly": 0.95,
                            "column": "passenger_count",
                            "min_value": 0.0,
                            "max_value": 6.0,
                        },
                        "result": {
                            "element_count": 100000,
                            "unexpected_count": 1,
                            "unexpected_percent": 0.001,
                            "partial_unexpected_list": [7.0],
                            "unexpected_percent_total": 0.001,
                            "unexpected_percent_nonmissing": 0.001,
                            "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
                            "partial_unexpected_index_list": [48422],
                        },
                    },
                ],
            },
            batch_id="default_pandas_datasource-#ephemeral_pandas_asset",
        ):
            actual = result.describe_dict()

        validation_results = actual.pop("validation_results")  # type: ignore[misc] # key is required
        assert len(validation_results) == 1

        expected = {
            "success": True,
            "statistics": {
                "evaluated_validations": 1,
                "success_percent": 100.0,
                "successful_validations": 1,
                "unsuccessful_validations": 0,
            },
        }
        assert actual == expected

    def _build_file_backed_checkpoint(
        self, tmp_path: pathlib.Path, actions: list[ValidationAction] | None = None
    ) -> Checkpoint:
        actions = actions or []
        with working_directory(tmp_path):
            context = gx.get_context(mode="file")

        ds = context.data_sources.add_pandas(self.datasource_name)
        csv_path = (
            pathlib.Path(__file__).parent.parent
            / "test_sets"
            / "quickstart"
            / "yellow_tripdata_sample_2022-01.csv"
        )
        assert csv_path.exists()
        asset = ds.add_csv_asset(self.asset_name, filepath_or_buffer=csv_path)

        batch_definition = asset.add_batch_definition(self.batch_definition_name)
        suite = context.suites.add(
            ExpectationSuite(
                name=self.suite_name,
                expectations=[
                    gxe.ExpectColumnValuesToBeBetween(
                        column=self.column_name, min_value=0, max_value=10
                    ),
                    gxe.ExpectColumnMeanToBeBetween(
                        column=self.column_name,
                        min_value=0,
                        max_value=1,
                    ),
                ],
            )
        )

        validation_definition = context.validation_definitions.add(
            ValidationDefinition(
                name=self.validation_definition_name, data=batch_definition, suite=suite
            )
        )

        return Checkpoint(
            name=self.checkpoint_name,
            validation_definitions=[validation_definition],
            actions=actions,
        )

    @pytest.mark.unit
    def test_checkpoint_result_does_not_contain_dataframe(
        self,
        tmp_path: pathlib.Path,
    ):
        df = pd.DataFrame({"passenger_count": [1, 2, 3, 4, 5]})

        context = gx.get_context(mode="ephemeral")
        ds = context.data_sources.add_pandas(self.datasource_name)
        asset = ds.add_dataframe_asset(name=self.asset_name)
        batch_definition = asset.add_batch_definition_whole_dataframe(self.batch_definition_name)

        suites = context.suites.add(
            ExpectationSuite(
                name=self.suite_name,
                expectations=[
                    gxe.ExpectColumnValuesToBeBetween(
                        column=self.column_name, min_value=0, max_value=6
                    )
                ],
            )
        )
        validation_definition = context.validation_definitions.add(
            ValidationDefinition(
                name=self.validation_definition_name, data=batch_definition, suite=suites
            )
        )
        checkpoint = Checkpoint(
            name=self.checkpoint_name,
            validation_definitions=[validation_definition],
            result_format="COMPLETE",
        )
        results = checkpoint.run(batch_parameters={"dataframe": df})
        assert len(results.run_results) == 1
        result = list(results.run_results.values())[0]
        assert result.meta["batch_parameters"]["dataframe"] == DATAFRAME_REPLACEMENT_STR
        assert (
            result.meta["active_batch_definition"]["batch_identifiers"]["dataframe"]
            == DATAFRAME_REPLACEMENT_STR
        )

    @pytest.mark.filesystem
    def test_checkpoint_run_result_shape(self, tmp_path: pathlib.Path):
        checkpoint = self._build_file_backed_checkpoint(tmp_path)

        result = checkpoint.run()

        assert result.success is False
        assert result.describe_dict() == {
            "success": False,
            "statistics": {
                "evaluated_validations": 1,
                "success_percent": 0.0,
                "successful_validations": 0,
                "unsuccessful_validations": 1,
            },
            "validation_results": [
                {
                    "expectations": [
                        {
                            "expectation_type": "expect_column_values_to_be_between",
                            "kwargs": {
                                "batch_id": f"{self.datasource_name}-{self.asset_name}",
                                "column": self.column_name,
                                "max_value": 10.0,
                                "min_value": 0.0,
                            },
                            "result": {
                                "element_count": 100000,
                                "missing_count": 0,
                                "missing_percent": 0.0,
                                "partial_unexpected_counts": [],
                                "partial_unexpected_index_list": [],
                                "partial_unexpected_list": [],
                                "unexpected_count": 0,
                                "unexpected_percent": 0.0,
                                "unexpected_percent_nonmissing": 0.0,
                                "unexpected_percent_total": 0.0,
                            },
                            "success": True,
                        },
                        {
                            "expectation_type": "expect_column_mean_to_be_between",
                            "kwargs": {
                                "batch_id": f"{self.datasource_name}-{self.asset_name}",
                                "column": self.column_name,
                                "max_value": 1.0,
                                "min_value": 0.0,
                            },
                            "result": {"observed_value": 1.54471},
                            "success": False,
                        },
                    ],
                    "statistics": {
                        "evaluated_expectations": 2,
                        "success_percent": 50.0,
                        "successful_expectations": 1,
                        "unsuccessful_expectations": 1,
                    },
                    "success": False,
                    "result_url": None,
                },
            ],
        }

    @pytest.mark.filesystem
    def test_checkpoint_run_adds_requisite_ids(
        self,
        tmp_path: pathlib.Path,
    ):
        checkpoint = self._build_file_backed_checkpoint(tmp_path)

        # A checkpoint that has not been persisted before running
        # should be saved during the run
        # This is also true for its nested validation definitions
        assert checkpoint.id is None

        result = checkpoint.run()

        assert checkpoint.id is not None

        # The meta attribute of each run result should contain the requisite IDs
        run_results = tuple(result.run_results.values())
        meta = run_results[0].meta
        assert sorted(meta.keys()) == [
            "active_batch_definition",
            "batch_markers",
            "batch_parameters",
            "batch_spec",
            "checkpoint_id",
            "great_expectations_version",
            "run_id",
            "validation_id",
            "validation_time",
        ]
        assert meta["checkpoint_id"] == checkpoint.id
        assert meta["validation_id"] == checkpoint.validation_definitions[0].id

    @pytest.mark.filesystem
    def test_checkpoint_run_with_data_docs_and_slack_actions_emit_page_links(
        self,
        tmp_path: pathlib.Path,
    ):
        actions = [
            SlackNotificationAction(name="slack_action", slack_webhook="webhook"),
            UpdateDataDocsAction(name="docs_action"),
        ]
        checkpoint = self._build_file_backed_checkpoint(tmp_path=tmp_path, actions=actions)

        with mock.patch.object(Session, "post") as mock_post:
            _ = checkpoint.run()

        mock_post.assert_called_once()
        docs_block = mock_post.call_args.kwargs["json"]["blocks"][3]["text"]["text"]
        assert "*DataDocs*" in docs_block
        assert "local_site" in docs_block
        assert "file:///" in docs_block

    @pytest.mark.filesystem
    def test_checkpoint_run_with_slack_action_no_page_links(
        self,
        tmp_path: pathlib.Path,
    ):
        actions = [
            SlackNotificationAction(name="slack_action", slack_webhook="webhook"),
        ]
        checkpoint = self._build_file_backed_checkpoint(tmp_path=tmp_path, actions=actions)

        with mock.patch.object(Session, "post") as mock_post:
            _ = checkpoint.run()

        mock_post.assert_called_once()
        blocks = mock_post.call_args.kwargs["json"]["blocks"]
        # No DataDocs blocks should be present
        assert blocks == [
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": "slack_action - my_checkpoint - Failure :no_entry:",
                },
            },
            {"type": "section", "text": {"type": "plain_text", "text": mock.ANY}},
            {
                "type": "section",
                "text": {
                    "text": "\n"
                    "*Asset*: `my_asset`  \n"
                    "*Expectation Suite*: `my_suite`\n"
                    "*Summary*: *1* of *2* Expectations were met",
                    "type": "mrkdwn",
                },
            },
            {"type": "divider"},
        ]


@pytest.mark.parametrize(
    "has_id,has_validation_def_id,has_suite_id,has_batch_def_id,error_list",
    [
        pytest.param(
            True,
            True,
            True,
            True,
            [],
            id="checkpoint_id|validation_id|suite_id|batch_def_id",
        ),
        pytest.param(
            True,
            True,
            True,
            False,
            [BatchDefinitionNotAddedError],
            id="checkpoint_id|validation_id|suite_id|no_batch_def_id",
        ),
        pytest.param(
            True,
            True,
            False,
            True,
            [ExpectationSuiteNotAddedError],
            id="checkpoint_id|validation_id|no_suite_id|batch_def_id",
        ),
        pytest.param(
            True,
            True,
            False,
            False,
            [BatchDefinitionNotAddedError, ExpectationSuiteNotAddedError],
            id="checkpoint_id|validation_id|no_suite_id|no_batch_def_id",
        ),
        pytest.param(
            True,
            False,
            True,
            True,
            [ValidationDefinitionNotAddedError],
            id="checkpoint_id|no_validation_id|suite_id|batch_def_id",
        ),
        pytest.param(
            True,
            False,
            True,
            False,
            [BatchDefinitionNotAddedError, ValidationDefinitionNotAddedError],
            id="checkpoint_id|no_validation_id|suite_id|no_batch_def_id",
        ),
        pytest.param(
            True,
            False,
            False,
            True,
            [ExpectationSuiteNotAddedError, ValidationDefinitionNotAddedError],
            id="checkpoint_id|no_validation_id|no_suite_id|batch_def_id",
        ),
        pytest.param(
            True,
            False,
            False,
            False,
            [
                BatchDefinitionNotAddedError,
                ExpectationSuiteNotAddedError,
                ValidationDefinitionNotAddedError,
            ],
            id="checkpoint_id|no_validation_id|no_suite_id|no_batch_def_id",
        ),
        pytest.param(
            False,
            True,
            True,
            True,
            [CheckpointNotAddedError],
            id="no_checkpoint_id|validation_id|suite_id|batch_def_id",
        ),
        pytest.param(
            False,
            True,
            True,
            False,
            [BatchDefinitionNotAddedError, CheckpointNotAddedError],
            id="no_checkpoint_id|validation_id|suite_id|no_batch_def_id",
        ),
        pytest.param(
            False,
            True,
            False,
            True,
            [ExpectationSuiteNotAddedError, CheckpointNotAddedError],
            id="no_checkpoint_id|validation_id|no_suite_id|batch_def_id",
        ),
        pytest.param(
            False,
            True,
            False,
            False,
            [BatchDefinitionNotAddedError, ExpectationSuiteNotAddedError, CheckpointNotAddedError],
            id="no_checkpoint_id|validation_id|no_suite_id|no_batch_def_id",
        ),
        pytest.param(
            False,
            False,
            True,
            True,
            [ValidationDefinitionNotAddedError, CheckpointNotAddedError],
            id="no_checkpoint_id|no_validation_id|suite_id|batch_def_id",
        ),
        pytest.param(
            False,
            False,
            True,
            False,
            [
                BatchDefinitionNotAddedError,
                ValidationDefinitionNotAddedError,
                CheckpointNotAddedError,
            ],
            id="no_checkpoint_id|no_validation_id|suite_id|no_batch_def_id",
        ),
        pytest.param(
            False,
            False,
            False,
            True,
            [
                ExpectationSuiteNotAddedError,
                ValidationDefinitionNotAddedError,
                CheckpointNotAddedError,
            ],
            id="no_checkpoint_id|no_validation_id|no_suite_id|batch_def_id",
        ),
        pytest.param(
            False,
            False,
            False,
            False,
            [
                BatchDefinitionNotAddedError,
                ExpectationSuiteNotAddedError,
                ValidationDefinitionNotAddedError,
                CheckpointNotAddedError,
            ],
            id="no_checkpoint_id|no_validation_id|no_suite_id|no_batch_def_id",
        ),
    ],
)
@pytest.mark.unit
def test_is_fresh(
    in_memory_runtime_context,
    has_id: bool,
    has_validation_def_id: bool,
    has_suite_id: bool,
    has_batch_def_id: bool,
    error_list: list[Type[ResourceFreshnessError]],
):
    context = in_memory_runtime_context

    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )

    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation_definition",
            suite=suite,
            data=batch_definition,
        )
    )

    checkpoint = context.checkpoints.add(
        Checkpoint(
            name="my_checkpoint",
            validation_definitions=[validation_definition],
        )
    )

    # Stores/Fluent API will always assign IDs but we manually override them here
    # for purposes of changing object state for the test
    if not has_batch_def_id:
        checkpoint.validation_definitions[0].data.id = None
    if not has_suite_id:
        checkpoint.validation_definitions[0].suite.id = None
    if not has_validation_def_id:
        checkpoint.validation_definitions[0].id = None
    if not has_id:
        checkpoint.id = None

    diagnostics = checkpoint.is_fresh()
    try:
        diagnostics.raise_for_error()
    except ResourceFreshnessAggregateError as e:
        assert [type(err) for err in e.errors] == error_list


@pytest.mark.unit
def test_is_fresh_raises_error_when_checkpoint_not_found(in_memory_runtime_context):
    context = in_memory_runtime_context

    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )

    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation_definition",
            suite=suite,
            data=batch_definition,
        )
    )

    checkpoint = context.checkpoints.add(
        Checkpoint(
            name="my_checkpoint",
            validation_definitions=[validation_definition],
        )
    )

    context.checkpoints.delete(checkpoint.name)

    diagnostics = checkpoint.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], CheckpointNotFoundError)


@pytest.mark.unit
def test_is_fresh_raises_error_when_child_deps_not_found(in_memory_runtime_context):
    context = in_memory_runtime_context

    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )

    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation_definition",
            suite=suite,
            data=batch_definition,
        )
    )

    checkpoint = context.checkpoints.add(
        Checkpoint(
            name="my_checkpoint",
            validation_definitions=[validation_definition],
        )
    )

    context.validation_definitions.delete(validation_definition.name)

    diagnostics = checkpoint.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], ValidationDefinitionNotFoundError)


@pytest.mark.unit
def test_is_fresh_raises_errors_for_all_child_validation_definitions(in_memory_runtime_context):
    context = in_memory_runtime_context

    datasource = context.data_sources.add_pandas(name="my_pandas_ds")
    asset = datasource.add_dataframe_asset(name="my_pandas_asset")
    bd_1 = asset.add_batch_definition_whole_dataframe(name="my_bd_1")
    bd_2 = asset.add_batch_definition_whole_dataframe(name="my_bd_2")

    suite_1 = ExpectationSuite(name="my_suite_1")
    suite_2 = ExpectationSuite(name="my_suite_2")

    vd_1 = ValidationDefinition(name="my_vd_1", data=bd_1, suite=suite_1)
    vd_2 = ValidationDefinition(name="my_vd_2", data=bd_2, suite=suite_2)

    cp = Checkpoint(name="my_cp", validation_definitions=[vd_1, vd_2])

    diagnostics = cp.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 5
    assert [type(err) for err in diagnostics.errors] == [
        ExpectationSuiteNotAddedError,
        ValidationDefinitionNotAddedError,
        ExpectationSuiteNotAddedError,
        ValidationDefinitionNotAddedError,
        CheckpointNotAddedError,
    ]


class TestCheckpointPydanticSerializationMethods:
    """
    Test overridden Pydantic serialization methods for Checkpoint
    (dict and json)
    """

    datasource_name: str = "my_pandas_ds"
    asset_name: str = "my_pandas_asset"
    batch_definition_name: str = "my_bd"
    suite_name: str = "my_suite"
    validation_definition_name: str = "my_vd"
    checkpoint_name: str = "my_cp"

    def _build_checkpoint_with_factories(
        self,
        context: AbstractDataContext,
    ) -> Checkpoint:
        datasource = context.data_sources.add_pandas(name=self.datasource_name)
        asset = datasource.add_dataframe_asset(name=self.asset_name)
        bd = asset.add_batch_definition_whole_dataframe(name=self.batch_definition_name)
        suite = context.suites.add(ExpectationSuite(name=self.suite_name))
        vd = context.validation_definitions.add(
            ValidationDefinition(name=self.validation_definition_name, data=bd, suite=suite)
        )
        return context.checkpoints.add(
            Checkpoint(name=self.checkpoint_name, validation_definitions=[vd])
        )

    def _build_checkpoint_without_factories(self) -> Checkpoint:
        bd = BatchDefinition(name=self.batch_definition_name)
        suite = ExpectationSuite(name=self.suite_name)
        vd = ValidationDefinition(name=self.validation_definition_name, data=bd, suite=suite)
        return Checkpoint(name=self.checkpoint_name, validation_definitions=[vd])

    @pytest.mark.unit
    def test_dict_serializes_correctly(self, in_memory_runtime_context):
        context = in_memory_runtime_context
        cp = self._build_checkpoint_with_factories(context)

        dict_val = cp.dict()
        assert dict_val == {
            "actions": [],
            "id": mock.ANY,
            "name": self.checkpoint_name,
            "result_format": "SUMMARY",
            "validation_definitions": [
                {
                    "id": mock.ANY,
                    "name": self.validation_definition_name,
                },
            ],
        }

        for id in (dict_val["id"], dict_val["validation_definitions"][0]["id"]):
            try:
                uuid.UUID(id)
            except TypeError:
                pytest.fail("id is not a valid UUID")

    @pytest.mark.unit
    def test_dict_raises_freshness_errors(self):
        cp = self._build_checkpoint_without_factories()
        with pytest.raises(CheckpointRelatedResourcesFreshnessError) as e:
            cp.dict()

        assert len(e.value.errors) == 3
        assert [type(err) for err in e.value.errors] == [
            BatchDefinitionNotAddedError,
            ExpectationSuiteNotAddedError,
            ValidationDefinitionNotAddedError,
        ]

    @pytest.mark.unit
    def test_json_serializes_correctly(self, in_memory_runtime_context):
        context = in_memory_runtime_context

        cp = self._build_checkpoint_with_factories(context)

        json_val = cp.json()
        dict_val = json.loads(json_val)
        assert dict_val == {
            "actions": [],
            "id": mock.ANY,
            "name": self.checkpoint_name,
            "result_format": "SUMMARY",
            "validation_definitions": [
                {
                    "id": mock.ANY,
                    "name": self.validation_definition_name,
                },
            ],
        }

        for id in (dict_val["id"], dict_val["validation_definitions"][0]["id"]):
            try:
                uuid.UUID(id)
            except TypeError:
                pytest.fail("id is not a valid UUID")

    @pytest.mark.unit
    def test_json_raises_freshness_errors(self):
        cp = self._build_checkpoint_without_factories()
        with pytest.raises(CheckpointRelatedResourcesFreshnessError) as e:
            cp.json()

        assert len(e.value.errors) == 3
        assert [type(err) for err in e.value.errors] == [
            BatchDefinitionNotAddedError,
            ExpectationSuiteNotAddedError,
            ValidationDefinitionNotAddedError,
        ]


@mock.patch(
    "great_expectations.data_context.data_context.context_factory.project_manager.is_using_cloud",
)
@pytest.mark.unit
def test_loaded_checkpoint_can_run(
    mock_is_using_cloud,
    empty_data_context: AbstractDataContext,
):
    mock_is_using_cloud.return_value = True
    col = "col"
    name = "checkpoint_testing"
    bd = (
        empty_data_context.data_sources.add_pandas(name)
        .add_dataframe_asset(name)
        .add_batch_definition_whole_dataframe(name)
    )
    suite = empty_data_context.suites.add(
        ExpectationSuite(
            name="test_suite",
            expectations=[
                gxe.ExpectColumnValuesToBeInSet(
                    column=col,
                    value_set=[1],
                )
            ],
        )
    )
    vd = empty_data_context.validation_definitions.add(
        ValidationDefinition(
            name=name,
            suite=suite,
            data=bd,
        )
    )
    empty_data_context.checkpoints.add(
        Checkpoint(
            name=name,
            validation_definitions=[vd],
        )
    )

    cp = empty_data_context.checkpoints.get(name)

    results = cp.run({"dataframe": pd.DataFrame({col: [1, 1]})})
    assert results.success


@pytest.mark.unit
def test_checkpoint_expectation_parameters(
    empty_data_context: AbstractDataContext,
) -> None:
    col = "col"
    name = "checkpoint_testing"
    bd = (
        empty_data_context.data_sources.add_pandas(name)
        .add_dataframe_asset(name)
        .add_batch_definition_whole_dataframe(name)
    )
    suite = empty_data_context.suites.add(
        ExpectationSuite(
            name="test_suite",
            expectations=[
                gxe.ExpectColumnValuesToBeInSet(
                    column=col,
                    value_set={"$PARAMETER": "values"},
                )
            ],
        )
    )
    vd = empty_data_context.validation_definitions.add(
        ValidationDefinition(
            name=name,
            suite=suite,
            data=bd,
        )
    )
    checkpoint = empty_data_context.checkpoints.add(
        Checkpoint(
            name=name,
            validation_definitions=[vd],
        )
    )

    results = checkpoint.run(
        expectation_parameters={"values": [1, 2]},
        batch_parameters={"dataframe": pd.DataFrame({col: [1, 2]})},
    )
    assert results.success


@pytest.mark.unit
def test_windowed_expectation_runs(
    empty_cloud_context_fluent: CloudDataContext,
    mocker: MockerFixture,
) -> None:
    # Arrange: Setup context and entities so we can run a checkpoint
    context = empty_cloud_context_fluent
    col_name = "my_col"
    dataframe = pd.DataFrame({col_name: [1, 2]})

    batch_def = (
        context.data_sources.add_pandas("my_datasource")
        .add_dataframe_asset("my_asset")
        .add_batch_definition("my_batch_definition")
    )
    expectation = gx.expectations.ExpectColumnMeanToBeBetween(
        column=col_name,
        min_value={"$PARAMETER": "my_min"},
        max_value={"$PARAMETER": "my_max"},
        windows=[
            Window(
                constraint_fn="mean",
                parameter_name="my_min",
                range=3,
                offset=Offset(positive=0.05, negative=0.05),
                strict=False,
            ),
            Window(
                constraint_fn="mean",
                parameter_name="my_max",
                range=3,
                offset=Offset(positive=0.05, negative=0.05),
                strict=False,
            ),
        ],
    )
    suite = context.suites.add(gx.ExpectationSuite("my_suite", expectations=[expectation]))
    validation_def = context.validation_definitions.add(
        gx.ValidationDefinition(name="my_validation_definition", suite=suite, data=batch_def)
    )
    checkpoint = context.checkpoints.add(
        Checkpoint(name="my_checkpoint", validation_definitions=[validation_def])
    )

    # Patch external requests made when running a checkpoint.
    # Since we are using mocker instead of mock, we don't need to use this as a context manager
    (
        mocker.patch.object(
            Checkpoint, "is_fresh", return_value=CheckpointFreshnessDiagnostics(errors=[])
        ),
    )
    # We set the returned min and max values returned by cloud
    # This is a magic mock since it needs to support being used as a context manager
    mock_session = mocker.MagicMock()
    cloud_min = 0.234
    cloud_max = 0.456
    mock_session.__enter__.return_value.get.return_value.json.return_value = {
        "data": {"expectation_parameters": {"my_min": cloud_min, "my_max": cloud_max}}
    }
    mocker.patch(
        "great_expectations.data_context.data_context.cloud_data_context.create_session",
        return_value=mock_session,
    )

    # Act: Run the checkpoint
    checkpoint_result = checkpoint.run(
        {
            "dataframe": dataframe,
        }
    )

    # Assert: Look at the configured expectation min and max and compare it to the returned cloud
    # values
    assert len(checkpoint_result.run_results) == 1
    outer_results = list(checkpoint_result.run_results.values())[0]
    result_list = outer_results.results
    assert len(result_list) == 1
    result = result_list[0]
    assert (
        result["expectation_config"]["rendered_content"][0].value.params["min_value"]["value"]
        == cloud_min
    )
    assert (
        result["expectation_config"]["rendered_content"][0].value.params["max_value"]["value"]
        == cloud_max
    )



from __future__ import annotations

import logging
import os
import pathlib
import uuid
from typing import TYPE_CHECKING, Final, Generator, Iterator

import numpy as np
import pandas as pd
import pytest

import great_expectations as gx
import great_expectations.exceptions as gx_exceptions
from great_expectations.checkpoint import Checkpoint
from great_expectations.core import ExpectationSuite
from great_expectations.data_context import CloudDataContext
from great_expectations.execution_engine import (
    SparkDFExecutionEngine,
)
from great_expectations.expectations import ExpectColumnValuesToNotBeNull

if TYPE_CHECKING:
    from great_expectations.compatibility import pyspark

LOGGER: Final = logging.getLogger("tests")


@pytest.fixture(scope="package")
def context() -> CloudDataContext:
    context = gx.get_context(
        mode="cloud",
        cloud_base_url=os.environ.get("GX_CLOUD_BASE_URL"),
        cloud_organization_id=os.environ.get("GX_CLOUD_ORGANIZATION_ID"),
        cloud_access_token=os.environ.get("GX_CLOUD_ACCESS_TOKEN"),
    )
    assert isinstance(context, CloudDataContext)
    return context


@pytest.fixture(scope="module")
def test_data() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "name": [1, 2, 3, 4],
        },
    )


@pytest.fixture(scope="module")
def datasource_name(
    context: CloudDataContext,
) -> Iterator[str]:
    datasource_name = f"ds_{uuid.uuid4().hex}"
    yield datasource_name
    # if the test was skipped, we may not have a datasource to clean up
    # in that case, we create one simply to test get and delete
    try:
        _ = context.data_sources.get(name=datasource_name)
    except KeyError:
        _ = context.data_sources.add_pandas(name=datasource_name)
        context.data_sources.get(name=datasource_name)
    context.delete_datasource(name=datasource_name)
    with pytest.raises(KeyError):
        _ = context.data_sources.get(name=datasource_name)


@pytest.fixture(scope="module")
def expectation_suite(
    context: CloudDataContext,
) -> Iterator[ExpectationSuite]:
    """This ExpectationSuite is shared by each E2E test, so its expected that the data
    used by each test follows the same shape."""
    expectation_suite_name = f"es_{uuid.uuid4().hex}"
    context.suites.add(
        ExpectationSuite(
            name=expectation_suite_name,
            expectations=[ExpectColumnValuesToNotBeNull(column="name", mostly=1)],
        )
    )
    yield context.suites.get(name=expectation_suite_name)
    context.suites.delete(expectation_suite_name)
    with pytest.raises(gx_exceptions.DataContextError):
        context.suites.get(name=expectation_suite_name)


@pytest.fixture(scope="module")
def checkpoint(
    context: CloudDataContext,
) -> Generator[Checkpoint, None, None]:
    """This Checkpoint is used by each E2E test. It's expected that each test
    will override its list of validation definitions within the test module."""
    checkpoint_name = f"E2E Test Checkpoint {uuid.uuid4().hex}"

    checkpoint = Checkpoint(name=checkpoint_name, validation_definitions=[])
    checkpoint = context.checkpoints.add(checkpoint=checkpoint)
    yield checkpoint
    context.checkpoints.delete(name=checkpoint_name)

    with pytest.raises(gx_exceptions.DataContextError):
        context.checkpoints.get(name=checkpoint_name)


@pytest.fixture(scope="module")
def tmp_path(tmp_path_factory) -> pathlib.Path:
    return tmp_path_factory.mktemp("project")


def construct_spark_df_from_pandas(
    spark_session: pyspark.SparkSession,
    pandas_df: pd.DataFrame,
) -> pyspark.DataFrame:
    spark_df = spark_session.createDataFrame(
        [
            tuple(
                None if isinstance(x, (float, int)) and np.isnan(x) else x for x in record.tolist()
            )
            for record in pandas_df.to_records(index=False)
        ],
        pandas_df.columns.tolist(),
    )
    return spark_df


@pytest.fixture(scope="module")
def spark_session() -> pyspark.SparkSession:
    from great_expectations.compatibility import pyspark

    if pyspark.SparkSession:  # type: ignore[truthy-function] # FIXME CoP
        return SparkDFExecutionEngine.get_or_create_spark_session()

    raise ValueError("spark tests are requested, but pyspark is not installed")



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Optional, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.analytics.client import submit as submit_event
from great_expectations.analytics.events import ValidationDefinitionRanEvent
from great_expectations.compatibility.pydantic import (
    BaseModel,
    Extra,
    ValidationError,
    validator,
)
from great_expectations.constants import DATAFRAME_REPLACEMENT_STR
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import (
    ExpectationSuite,
)
from great_expectations.core.freshness_diagnostics import (
    ValidationDefinitionFreshnessDiagnostics,
)
from great_expectations.core.result_format import DEFAULT_RESULT_FORMAT
from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.core.serdes import _EncodedValidationData, _IdentifierBundle
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.data_context.types.refs import GXCloudResourceRef
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    GXCloudIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.exceptions import (
    ValidationDefinitionNotAddedError,
    ValidationDefinitionNotFreshError,
)
from great_expectations.exceptions.exceptions import (
    InvalidKeyError,
    StoreBackendError,
    ValidationDefinitionNotFoundError,
)
from great_expectations.validator.v1_validator import Validator

if TYPE_CHECKING:
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )
    from great_expectations.core.result_format import ResultFormatUnion
    from great_expectations.core.suite_parameters import SuiteParameterDict
    from great_expectations.data_context.store.validation_results_store import (
        ValidationResultsStore,
    )
    from great_expectations.datasource.fluent.batch_request import BatchParameters
    from great_expectations.datasource.fluent.interfaces import DataAsset, Datasource


@public_api
class ValidationDefinition(BaseModel):
    """
    Responsible for running a suite against data and returning a validation result.

    Args:
        name: The name of the validation.
        data: A batch definition to validate.
        suite: A grouping of expectations to validate against the data.
        id: A unique identifier for the validation; added when persisted with a store.

    """

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True  # Necessary for compatibility with suite's Marshmallow dep
        copy_on_model_validation = (
            "none"  # Necessary to prevent cloning when passing to a checkpoint
        )
        validate_assignment = True
        """
        When serialized, the suite and data fields should be encoded as a set of identifiers.
        These will be used as foreign keys to retrieve the actual objects from the appropriate stores.

        Example:
        {
            "name": "my_validation",
            "data": {
                "datasource": {
                    "name": "my_datasource",
                    "id": "a758816-64c8-46cb-8f7e-03c12cea1d67"
                },
                "asset": {
                    "name": "my_asset",
                    "id": "b5s8816-64c8-46cb-8f7e-03c12cea1d67"
                },
                "batch_definition": {
                    "name": "my_batch_definition",
                    "id": "3a758816-64c8-46cb-8f7e-03c12cea1d67"
                }
            },
            "suite": {
                "name": "my_suite",
                "id": "8r2g816-64c8-46cb-8f7e-03c12cea1d67"
            },
            "id": "20dna816-64c8-46cb-8f7e-03c12cea1d67"
        }
        """  # noqa: E501 # FIXME CoP
        json_encoders = {
            ExpectationSuite: lambda e: e.identifier_bundle(),
            BatchDefinition: lambda b: b.identifier_bundle(),
        }

    name: str
    data: BatchDefinition
    suite: ExpectationSuite
    id: Union[str, None] = None

    @property
    @public_api
    def batch_definition(self) -> BatchDefinition:
        """
        The Batch Definition to validate.
        """
        return self.data

    @property
    @public_api
    def asset(self) -> DataAsset:
        """
        The parent Data Asset of the Batch Definition.
        """
        return self.data.data_asset

    @property
    def data_source(self) -> Datasource:
        return self.asset.datasource

    @property
    def _validation_results_store(self) -> ValidationResultsStore:
        return project_manager.get_validation_results_store()

    def is_fresh(self) -> ValidationDefinitionFreshnessDiagnostics:
        validation_definition_diagnostics = ValidationDefinitionFreshnessDiagnostics(
            errors=[] if self.id else [ValidationDefinitionNotAddedError(name=self.name)]
        )
        suite_diagnostics = self.suite.is_fresh()
        data_diagnostics = self.data.is_fresh()
        validation_definition_diagnostics.update_with_children(suite_diagnostics, data_diagnostics)

        if not validation_definition_diagnostics.success:
            return validation_definition_diagnostics

        store = project_manager.get_validation_definition_store()
        key = store.get_key(name=self.name, id=self.id)

        try:
            validation_definition = store.get(key=key)
        except (
            StoreBackendError,  # Generic error from stores
            InvalidKeyError,  # Ephemeral context error
        ):
            return ValidationDefinitionFreshnessDiagnostics(
                errors=[ValidationDefinitionNotFoundError(name=self.name)]
            )

        return ValidationDefinitionFreshnessDiagnostics(
            errors=[]
            if self == validation_definition
            else [ValidationDefinitionNotFreshError(name=self.name)]
        )

    @validator("suite", pre=True)
    def _validate_suite(cls, v: dict | ExpectationSuite):
        # Input will be a dict of identifiers if being deserialized or a suite object if being constructed by a user.  # noqa: E501 # FIXME CoP
        if isinstance(v, dict):
            return cls._decode_suite(v)
        elif isinstance(v, ExpectationSuite):
            return v
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            "Suite must be a dictionary (if being deserialized) or an ExpectationSuite object."
        )

    @validator("data", pre=True)
    def _validate_data(cls, v: dict | BatchDefinition):
        # Input will be a dict of identifiers if being deserialized or a rich type if being constructed by a user.  # noqa: E501 # FIXME CoP
        if isinstance(v, dict):
            return cls._decode_data(v)
        elif isinstance(v, BatchDefinition):
            return v
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            "Data must be a dictionary (if being deserialized) or a BatchDefinition object."
        )

    @classmethod
    def _decode_suite(cls, suite_dict: dict) -> ExpectationSuite:
        # Take in raw JSON, ensure it contains appropriate identifiers, and use them to retrieve the actual suite.  # noqa: E501 # FIXME CoP
        try:
            suite_identifiers = _IdentifierBundle.parse_obj(suite_dict)
        except ValidationError as e:
            raise ValueError("Serialized suite did not contain expected identifiers") from e  # noqa: TRY003 # FIXME CoP

        name = suite_identifiers.name
        id = suite_identifiers.id

        expectation_store = project_manager.get_expectations_store()
        key = expectation_store.get_key(name=name, id=id)

        try:
            config: dict = expectation_store.get(key)
        except gx_exceptions.InvalidKeyError as e:
            raise ValueError(f"Could not find suite with name: {name} and id: {id}") from e  # noqa: TRY003 # FIXME CoP

        suite = ExpectationSuite(**config)
        if suite._include_rendered_content:
            suite.render()
        return suite

    @classmethod
    def _decode_data(cls, data_dict: dict) -> BatchDefinition:
        # Take in raw JSON, ensure it contains appropriate identifiers, and use them to retrieve the actual data.  # noqa: E501 # FIXME CoP
        try:
            data_identifiers = _EncodedValidationData.parse_obj(data_dict)
        except ValidationError as e:
            raise ValueError("Serialized data did not contain expected identifiers") from e  # noqa: TRY003 # FIXME CoP

        ds_name = data_identifiers.datasource.name
        asset_name = data_identifiers.asset.name
        batch_definition_name = data_identifiers.batch_definition.name

        datasource_dict = project_manager.get_datasources()
        try:
            ds = datasource_dict[ds_name]
        except KeyError as e:
            raise ValueError(f"Could not find datasource named '{ds_name}'.") from e  # noqa: TRY003 # FIXME CoP

        try:
            asset = ds.get_asset(asset_name)
        except LookupError as e:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"Could not find asset named '{asset_name}' within '{ds_name}' datasource."
            ) from e

        try:
            batch_definition = asset.get_batch_definition(batch_definition_name)
        except KeyError as e:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"Could not find batch definition named '{batch_definition_name}' within '{asset_name}' asset and '{ds_name}' datasource."  # noqa: E501 # FIXME CoP
            ) from e

        return batch_definition

    @public_api
    def run(
        self,
        *,
        checkpoint_id: Optional[str] = None,
        batch_parameters: Optional[BatchParameters] = None,
        expectation_parameters: Optional[SuiteParameterDict] = None,
        result_format: ResultFormatUnion = DEFAULT_RESULT_FORMAT,
        run_id: RunIdentifier | None = None,
    ) -> ExpectationSuiteValidationResult:
        """
        Runs a validation using the configured data and suite.

        Args:
            batch_parameters: The dictionary of parameters necessary for selecting the
              correct batch to run the validation on. The keys are strings that are determined
              by the BatchDefinition used to instantiate this ValidationDefinition. For example:
              - whole table -> None
              - yearly -> year
              - monthly -> year, month
              - daily -> year, month, day

            expectation_parameters: A dictionary of parameters values for any expectations using
              parameterized values (the $PARAMETER syntax). The keys are the parameter names
              and the values are the values to be used for this validation run.
            result_format: A parameter controlling how much diagnostic information the result
              contains.
            checkpoint_id: This is used by the checkpoints code when it runs a validation
              definition. Otherwise, it should be None.
            run_id: An identifier for this run. Typically, this should be set to None and it will
              be generated by this call.
        """
        diagnostics = self.is_fresh()
        if not diagnostics.success:
            # The validation definition itself is not added but all children are - we can add it for the user # noqa: E501 # FIXME CoP
            if not diagnostics.parent_added and diagnostics.children_added:
                self._add_to_store()
            else:
                diagnostics.raise_for_error()

        validator = Validator(
            batch_definition=self.batch_definition,
            batch_parameters=batch_parameters,
            result_format=result_format,
        )
        results = validator.validate_expectation_suite(self.suite, expectation_parameters)
        results.meta["validation_id"] = self.id
        results.meta["checkpoint_id"] = checkpoint_id

        # NOTE: We should promote this to a top-level field of the result.
        #       Meta should be reserved for user-defined information.
        if not run_id:
            run_id = RunIdentifier(run_time=datetime.datetime.now(datetime.timezone.utc))
        results.meta["run_id"] = run_id
        results.meta["validation_time"] = run_id.run_time

        if batch_parameters:
            batch_parameters_copy = {k: v for k, v in batch_parameters.items()}
            if "dataframe" in batch_parameters_copy:
                batch_parameters_copy["dataframe"] = DATAFRAME_REPLACEMENT_STR
            results.meta["batch_parameters"] = batch_parameters_copy
        else:
            results.meta["batch_parameters"] = None

        (
            expectation_suite_identifier,
            validation_result_id,
        ) = self._get_expectation_suite_and_validation_result_ids(
            validator=validator, run_id=run_id
        )

        ref = self._validation_results_store.store_validation_results(
            suite_validation_result=results,
            suite_validation_result_identifier=validation_result_id,
            expectation_suite_identifier=expectation_suite_identifier,
        )

        if isinstance(ref, GXCloudResourceRef):
            results.id = ref.id
            # FIXME(cdkini): There is currently a bug in GX Cloud where the result_url is None
            results.result_url = self._validation_results_store.parse_result_url_from_gx_cloud_ref(
                ref
            )

        submit_event(
            event=ValidationDefinitionRanEvent(
                validation_definition_id=self.id,
                checkpoint_id=checkpoint_id,
            )
        )

        return results

    def _get_expectation_suite_and_validation_result_ids(
        self,
        validator: Validator,
        run_id: RunIdentifier | None = None,
    ) -> (
        tuple[GXCloudIdentifier, GXCloudIdentifier]
        | tuple[ExpectationSuiteIdentifier, ValidationResultIdentifier]
    ):
        expectation_suite_identifier: GXCloudIdentifier | ExpectationSuiteIdentifier
        validation_result_id: GXCloudIdentifier | ValidationResultIdentifier
        if self._validation_results_store.cloud_mode:
            expectation_suite_identifier = GXCloudIdentifier(
                resource_type=GXCloudRESTResource.EXPECTATION_SUITE,
                id=self.suite.id,
            )
            validation_result_id = GXCloudIdentifier(
                resource_type=GXCloudRESTResource.VALIDATION_RESULT
            )
            return expectation_suite_identifier, validation_result_id
        else:
            run_id = run_id or RunIdentifier(
                run_time=datetime.datetime.now(tz=datetime.timezone.utc)
            )
            expectation_suite_identifier = ExpectationSuiteIdentifier(name=self.suite.name)
            validation_result_id = ValidationResultIdentifier(
                batch_identifier=validator.active_batch_id,
                expectation_suite_identifier=expectation_suite_identifier,
                run_id=run_id,
            )
            return expectation_suite_identifier, validation_result_id

    def identifier_bundle(self) -> _IdentifierBundle:
        # Utilized as a custom json_encoder
        diagnostics = self.is_fresh()
        diagnostics.raise_for_error()

        return _IdentifierBundle(name=self.name, id=self.id)

    @public_api
    def save(self) -> None:
        """Save the current state of this ValidationDefinition."""
        store = project_manager.get_validation_definition_store()
        key = store.get_key(name=self.name, id=self.id)

        store.update(key=key, value=self)

    def _add_to_store(self) -> None:
        """This is used to persist a validation_definition before we run it.

        We need to persist a validation_definition before it can be run. If user calls runs but
        hasn't persisted it we add it for them."""
        store = project_manager.get_validation_definition_store()
        key = store.get_key(name=self.name, id=self.id)

        store.add(key=key, value=self)



from __future__ import annotations

import pytest

from great_expectations.compatibility.pyspark import types as pyspark_types
from great_expectations.datasource.fluent.serializable_types.pyspark import (
    SerializableStructType,
)


@pytest.fixture
def skip_if_spark_not_selected(test_backends) -> None:
    # NOTE: This should be edited when we organize tests to use marks instead of skips.
    if "SparkDFDataset" not in test_backends:
        pytest.skip("No spark backend selected.")


if pyspark_types:
    struct_type_test_params = [
        pytest.param(None, {"fields": [], "type": "struct"}, id="None"),
        pytest.param([], {"fields": [], "type": "struct"}, id="empty_list"),
        pytest.param(
            [pyspark_types.StructField("f1", pyspark_types.StringType(), True)],
            {
                "fields": [{"metadata": {}, "name": "f1", "nullable": True, "type": "string"}],
                "type": "struct",
            },
            id="fields_list",
        ),
        pytest.param(
            pyspark_types.StructType(
                [pyspark_types.StructField("f1", pyspark_types.StringType(), True)]
            ),
            {
                "fields": [{"metadata": {}, "name": "f1", "nullable": True, "type": "string"}],
                "type": "struct",
            },
            id="struct_type",
        ),
    ]
else:
    struct_type_test_params = []


@pytest.mark.unit
@pytest.mark.parametrize("fields_or_struct_type,serialized_output", struct_type_test_params)
def test_serializable_struct_type(
    fields_or_struct_type: pyspark_types.StructType | list[pyspark_types.StructField] | None,
    serialized_output: dict,
    skip_if_spark_not_selected: None,
):
    sst = SerializableStructType(fields_or_struct_type=fields_or_struct_type)
    assert isinstance(sst.struct_type, pyspark_types.StructType)
    assert isinstance(sst, dict)
    assert sst == serialized_output


@pytest.mark.unit
@pytest.mark.parametrize(
    "fields_or_struct_type,serialized_output",
    struct_type_test_params,
)
def test_serializable_struct_type_validate(
    fields_or_struct_type: pyspark_types.StructType | list[pyspark_types.StructField] | None,
    serialized_output: dict,
    skip_if_spark_not_selected: None,
):
    sst = SerializableStructType.validate(fields_or_struct_type)
    assert isinstance(sst.struct_type, pyspark_types.StructType)
    assert isinstance(sst, dict)
    assert sst == serialized_output



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Literal, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_DESCRIPTION,
    MOSTLY_DESCRIPTION,
)
from great_expectations.render.components import (
    LegacyRendererType,
    RenderedStringTemplateContent,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import num_to_str, substitute_none_for_missing

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationConfiguration,
    )
    from great_expectations.core.expectation_validation_result import (
        ExpectationValidationResult,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the column entries to be strings that match any "
    "of a provided list of like pattern expressions."
)
LIKE_PATTERN_DESCRIPTION = (
    "The list of SQL like pattern expressions the column entries should match."
)
MATCH_ON_DESCRIPTION = (
    "'any' or 'all'. "
    "Use 'any' if the value should match at least one like pattern in the list. "
    "Use 'all' if it should match each like pattern in the list."
)
DATA_QUALITY_ISSUES = [DataQualityIssues.VALIDITY.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.REDSHIFT.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
]


class ExpectColumnValuesToMatchLikePatternList(ColumnMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnValuesToMatchLikePatternList is a \
    Column Map Expectation.

    Column Map Expectations are one of the most common types of Expectation.
    They are evaluated for a single column and ask a yes/no question for every row in that column.
    Based on the result, they then calculate the percentage of rows that gave a positive answer. If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        like_pattern_list (List[str]): \
            {LIKE_PATTERN_DESCRIPTION}
        match_on (string): \
            {MATCH_ON_DESCRIPTION}

    Other Parameters:
        mostly (None or a float between 0 and 1): \
            {MOSTLY_DESCRIPTION} \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See Also:
        [ExpectColumnValuesToMatchRegex](https://greatexpectations.io/expectations/expect_column_values_to_match_regex)
        [ExpectColumnValuesToMatchRegexList](https://greatexpectations.io/expectations/expect_column_values_to_match_regex_list)
        [ExpectColumnValuesToNotMatchRegex](https://greatexpectations.io/expectations/expect_column_values_to_not_match_regex)
        [ExpectColumnValuesToNotMatchRegexList](https://greatexpectations.io/expectations/expect_column_values_to_not_match_regex_list)
        [ExpectColumnValuesToMatchLikePattern](https://greatexpectations.io/expectations/expect_column_values_to_match_like_pattern)
        [ExpectColumnValuesToNotMatchLikePattern](https://greatexpectations.io/expectations/expect_column_values_to_not_match_like_pattern)
        [ExpectColumnValuesToNotMatchLikePatternList](https://greatexpectations.io/expectations/expect_column_values_to_not_match_like_pattern_list)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	"aaa"   "ade"
            1 	"abb"   "adb"
            2 	"acc"   "aaa"

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnValuesToMatchLikePatternList(
                    column="test",
                    like_pattern_list=["[aa]%", "[ab]%", "[ac]%"],
                    match_on="any"
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 0,
                    "unexpected_percent": 0.0,
                    "partial_unexpected_list": [],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 0.0,
                    "unexpected_percent_nonmissing": 0.0
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnValuesToMatchLikePatternList(
                    column="test2",
                    like_pattern_list=["[ad]%", "[a]%"],
                    match_on="all"
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 1,
                    "unexpected_percent": 33.33333333333333,
                    "partial_unexpected_list": [
                      "aaa",
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 33.33333333333333,
                    "unexpected_percent_nonmissing": 33.33333333333333
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    like_pattern_list: Union[List[str], SuiteParameterDict] = pydantic.Field(
        description=LIKE_PATTERN_DESCRIPTION
    )
    match_on: Union[Literal["any", "all"], SuiteParameterDict] = pydantic.Field(
        default="any", description=MATCH_ON_DESCRIPTION
    )

    @pydantic.validator("like_pattern_list")
    def validate_like_pattern_list(
        cls, like_pattern_list: list[str] | SuiteParameterDict
    ) -> list[str] | SuiteParameterDict:
        if len(like_pattern_list) < 1:
            raise ValueError("At least one like_pattern must be supplied in the like_pattern_list.")  # noqa: TRY003 # FIXME CoP

        return like_pattern_list

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": [
            "@great_expectations",
        ],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "column_values.match_like_pattern_list"
    success_keys = ("mostly", "like_pattern_list", "match_on")
    args_keys = (
        "column",
        "like_pattern_list",
    )

    class Config:
        title = "Expect column values to match like pattern list"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnValuesToMatchLikePatternList]
        ) -> None:
            ColumnMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("like_pattern_list", RendererValueType.ARRAY),
            ("match_on", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if renderer_configuration.include_column_name:
            template_str = "$column values "
        else:
            template_str = "Values "

        if params.match_on:
            template_str += "must match $match_on of "
        else:
            template_str += "must match "

        if params.like_pattern_list:
            array_param_name = "like_pattern_list"
            param_prefix = "like_pattern_list_"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            template_str += "the following like patterns: " + cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )

        if params.mostly and params.mostly.value < 1.0:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            template_str += " , at least $mostly_pct % of the time."

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> List[RenderedStringTemplateContent]:
        runtime_configuration = runtime_configuration or {}
        _ = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")

        params = substitute_none_for_missing(
            configuration.kwargs,
            ["column", "like_pattern_list", "mostly"],
        )
        if params["mostly"] is not None:
            params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)

        if not params.get("like_pattern_list") or len(params.get("like_pattern_list")) == 0:
            values_string = "[ ]"
        else:
            for i, v in enumerate(params["like_pattern_list"]):
                params[f"v__{i!s}"] = v
            values_string = " ".join(
                [f"$v__{i!s}" for i, v in enumerate(params["like_pattern_list"])]
            )

        template_str = "Values must match the following like patterns: " + values_string

        if params["mostly"] is not None:
            if isinstance(params["mostly"], (int, float)) and params["mostly"] < 1.0:
                params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
                # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")  # noqa: E501 # FIXME CoP
            template_str += ", at least $mostly_pct % of the time."
        else:
            template_str += "."

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]



import pytest

from great_expectations.data_context.types.base import ProgressBarsConfig
from great_expectations.data_context.util import instantiate_class_from_config


@pytest.mark.unit
def test_instantiate_class_from_config():
    # This config structure feels very tidy to me.
    instantiate_class_from_config(
        config={
            "module_name": "tests.test_plugins.fake_configs",
            "class_name": "FakeConfigurableClass",
            "a": "hi",
        },
        runtime_environment={"x": 1},
    )


@pytest.mark.unit
def test_instantiate_class_from_config_with_overriden_defaults():
    # This config structure feels very tidy to me, even with nesting.
    fake_configurable_wrapper_object = instantiate_class_from_config(
        config={
            "module_name": "tests.test_plugins.fake_configs",
            "class_name": "FakeConfigurableWrapperClass",
            "foo": 100,
            "fake_configurable": {
                "module_name": "tests.test_plugins.fake_configs",
                "class_name": "FakeConfigurableClass",
            },
        },
        runtime_environment={"x": 1},
    )
    assert fake_configurable_wrapper_object.fake_configurable_object.a == "default_value_for_a"

    fake_configurable_wrapper_object = instantiate_class_from_config(
        config={
            "module_name": "tests.test_plugins.fake_configs",
            "class_name": "FakeConfigurableWrapperClass",
            "foo": 100,
            "fake_configurable": {
                "module_name": "tests.test_plugins.fake_configs",
                "class_name": "FakeConfigurableClass",
                "a": "not_the_default_value",
            },
        },
        runtime_environment={"x": 1},
    )
    assert fake_configurable_wrapper_object.fake_configurable_object.a == "not_the_default_value"


@pytest.mark.unit
def test_instantiate_class_from_config_with_config_defaults():
    my_runtime_environment = {"x": 1}

    # `a` specified in config only. No default
    fake_configurable_object = instantiate_class_from_config(
        config={
            "class_name": "FakeConfigurableClass",
            "module_name": "tests.test_plugins.fake_configs",
            "a": "value_from_the_config",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={},
    )
    assert fake_configurable_object.a == "value_from_the_config"

    # Default only. No config
    fake_configurable_object = instantiate_class_from_config(
        config={
            "class_name": "FakeConfigurableClass",
            "module_name": "tests.test_plugins.fake_configs",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={
            "a": "value_from_the_defaults",
        },
    )
    assert fake_configurable_object.a == "value_from_the_defaults"

    # Both.
    fake_configurable_object = instantiate_class_from_config(
        config={
            "class_name": "FakeConfigurableClass",
            "module_name": "tests.test_plugins.fake_configs",
            "a": "value_from_the_config",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={
            "a": "value_from_the_defaults",
        },
    )
    assert fake_configurable_object.a == "value_from_the_config"

    # Neither
    with pytest.raises(TypeError):
        fake_configurable_object = instantiate_class_from_config(
            config={
                "class_name": "FakeConfigurableClass",
                "module_name": "tests.test_plugins.fake_configs",
            },
            runtime_environment=my_runtime_environment,
            config_defaults={},
        )

    # Module name specified in default, but not config.
    fake_configurable_object = instantiate_class_from_config(
        config={
            "class_name": "FakeConfigurableClass",
            "a": "value_from_the_config",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={
            "module_name": "tests.test_plugins.fake_configs",
        },
    )

    # Both
    fake_configurable_object = instantiate_class_from_config(
        config={
            "module_name": "tests.test_plugins.fake_configs",
            "class_name": "FakeConfigurableClass",
            "a": "value_from_the_config",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={
            "module_name": "tests.test_plugins.fake_configs",
        },
    )

    # Neither
    with pytest.raises(KeyError):
        fake_configurable_object = instantiate_class_from_config(
            config={
                "class_name": "FakeConfigurableClass",
                "a": "value_from_the_config",
            },
            runtime_environment=my_runtime_environment,
            config_defaults={},
        )

    # Pushing the limits of what we can do with this API...
    fake_configurable_object = instantiate_class_from_config(
        config={
            "class_name": "FakeConfigurableClass",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={
            "module_name": "tests.test_plugins.fake_configs",
            "a": "value_from_the_config",
        },
    )

    # This seems like too much magic, but maybe we'll find a place where it's needed.
    fake_configurable_object = instantiate_class_from_config(
        config={
            "a": "value_from_the_config",
        },
        runtime_environment=my_runtime_environment,
        config_defaults={
            "module_name": "tests.test_plugins.fake_configs",
            "class_name": "FakeConfigurableClass",
        },
    )

    # Okay, this is just getting silly
    fake_configurable_object = instantiate_class_from_config(
        config={},
        runtime_environment=my_runtime_environment,
        config_defaults={
            "module_name": "tests.test_plugins.fake_configs",
            "class_name": "FakeConfigurableClass",
            "a": "value_from_the_config",
        },
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "progress_bars_config,expect_global,expect_metric_calculations",
    [
        (ProgressBarsConfig(), True, True),
        (ProgressBarsConfig(metric_calculations=False), True, False),
        (
            ProgressBarsConfig(metric_calculations=False),
            True,
            False,
        ),
        (
            ProgressBarsConfig(globally=True, metric_calculations=False),
            True,
            False,
        ),
        (
            ProgressBarsConfig(globally=False, metric_calculations=True),
            False,
            True,
        ),
        (ProgressBarsConfig(globally=False), False, False),
    ],
)
def test_config_pbar_option(
    progress_bars_config: ProgressBarsConfig,
    expect_global: bool,
    expect_metric_calculations: bool,
):
    enabled_pbars = [
        progress_bars_config.globally,
        progress_bars_config.metric_calculations,
    ]
    assert enabled_pbars == [
        expect_global,
        expect_metric_calculations,
    ]



from great_expectations.data_context.data_context import (
    AbstractDataContext,
    CloudDataContext,
    EphemeralDataContext,
    FileDataContext,
    get_context,
    project_manager,
    set_context,
)



# Import MapMetricProvider classes
from great_expectations.expectations.metrics.map_metric_provider.column_condition_partial import (
    column_condition_partial,
)

# Import decorators
from great_expectations.expectations.metrics.map_metric_provider.column_function_partial import (
    column_function_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.column_map_metric_provider import (
    ColumnMapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.column_pair_condition_partial import (  # noqa: E501 # FIXME CoP
    column_pair_condition_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.column_pair_function_partial import (  # noqa: E501 # FIXME CoP
    column_pair_function_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.column_pair_map_metric_provider import (  # noqa: E501 # FIXME CoP
    ColumnPairMapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.map_metric_provider import (
    MapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_condition_partial import (  # noqa: E501 # FIXME CoP
    multicolumn_condition_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_function_partial import (  # noqa: E501 # FIXME CoP
    multicolumn_function_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_map_metric_provider import (  # noqa: E501 # FIXME CoP
    MulticolumnMapMetricProvider,
)
